Package: rsi
Title: Efficiently Retrieve and Process Satellite Imagery
Version: 0.3.3
Authors@R: c(
    person("Michael", "Mahoney", , "mike.mahoney.218@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-2402-304X")),
        person("Felipe", "Carvalho", role = "rev",
           comment = "Felipe reviewed the package (v. 0.3.0) for rOpenSci, see <https://github.com/ropensci/software-review/issues/636>"),
    person("Michael", "Sumner", role = "rev",
           comment = "Michael reviewed the package (v. 0.3.0) for rOpenSci, see <https://github.com/ropensci/software-review/issues/636>"),
    person("Permian Global", role = c("cph", "fnd"))
  )
Description: Downloads spatial data from spatiotemporal asset catalogs 
    ('STAC'), computes standard spectral indices from the Awesome Spectral 
    Indices project (Montero et al. (2023) <doi:10.1038/s41597-023-02096-0>) 
    against raster data, and glues the outputs together into predictor bricks. 
    Methods focus on interoperability with the broader spatial ecosystem; 
    function arguments and outputs use classes from 'sf' and 'terra', and data 
    downloading functions support complex 'CQL2' queries using 'rstac'.
License: Apache License (>= 2)
Depends: R (>= 4.1)
Imports: future.apply, glue, httr, jsonlite, lifecycle,
        proceduralnames, rlang, rstac, sf, terra, tibble
Suggests: curl, knitr, progressr, rmarkdown, testthat (>= 3.0.0), withr
Config/testthat/edition: 3
Config/testthat/parallel: true
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.2
URL: https://github.com/Permian-Global-Research/rsi,
        https://permian-global-research.github.io/rsi/
BugReports: https://github.com/Permian-Global-Research/rsi/issues
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-09 00:04:09 UTC; mikemahoney218
Author: Michael Mahoney [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2402-304X>),
  Felipe Carvalho [rev] (Felipe reviewed the package (v. 0.3.0) for
    rOpenSci, see
    <https://github.com/ropensci/software-review/issues/636>),
  Michael Sumner [rev] (Michael reviewed the package (v. 0.3.0) for
    rOpenSci, see
    <https://github.com/ropensci/software-review/issues/636>),
  Permian Global [cph, fnd]
Maintainer: Michael Mahoney <mike.mahoney.218@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-09 09:50:02 UTC
Built: R 4.4.3; ; 2026-04-09 13:38:16 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",rsi_band_mapping)
S3method("[[",rsi_band_mapping)
S3method(c,rsi_band_mapping)
S3method(print,rsi_band_mapping)
export(alos_palsar_mask_function)
export(calculate_indices)
export(default_query_function)
export(filter_bands)
export(filter_platforms)
export(get_alos_palsar_imagery)
export(get_dem)
export(get_landsat_imagery)
export(get_naip_imagery)
export(get_sentinel1_imagery)
export(get_sentinel2_imagery)
export(get_stac_data)
export(landsat_mask_function)
export(landsat_platform_filter)
export(rsi_download_rasters)
export(rsi_gdal_config_options)
export(rsi_gdalwarp_options)
export(rsi_query_api)
export(sentinel2_mask_function)
export(sign_planetary_computer)
export(spectral_indices)
export(spectral_indices_url)
export(stack_rasters)
importFrom(httr,user_agent)
importFrom(rlang,"%||%")
importFrom(stats,predict)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# rsi <a href="https://permian-global-research.github.io/rsi/"><img src="man/figures/logo.png" align="right" height="138" alt="rsi website" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/Permian-Global-Research/rsi/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/Permian-Global-Research/rsi/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/Permian-Global-Research/rsi/branch/main/graph/badge.svg)](https://app.codecov.io/gh/Permian-Global-Research/rsi?branch=main)
[![License: Apache
2.0](https://img.shields.io/badge/License-Apache_2.0-blue.svg)](https://opensource.org/license/apache-2-0)
[![Lifecycle:
maturing](https://img.shields.io/badge/lifecycle-maturing-blue.svg)](https://lifecycle.r-lib.org/articles/stages.html#maturing)
[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![CRAN
status](https://www.r-pkg.org/badges/version/rsi)](https://CRAN.R-project.org/package=rsi)
[![DOI](https://zenodo.org/badge/682237263.svg)](https://zenodo.org/doi/10.5281/zenodo.10926857)
[![Status at rOpenSci Software Peer
Review](https://badges.ropensci.org/636_status.svg)](https://github.com/ropensci/software-review/issues/636)
<!-- badges: end -->

The goal of rsi is to address several **r**epeated **s**patial
**i**nfelicities, by providing utility functions that save you typing
and help avoid **r**epetitive **s**tress **i**njuries. Specifically, rsi
provides:

- An interface to the **R**some – excuse me, [*Awesome* Spectral Indices
  project](https://github.com/awesome-spectral-indices/awesome-spectral-indices),
  providing the list of indices directly in R as a friendly tibble,
- A method for efficiently *calculating* those awesome spectral indices
  using local rasters, enabling **r**apid **s**pectral **i**nference,
- A method for downloading STAC data – excuse me, **r**etriving **S**TAC
  **i**nformation – from any STAC server, with additional helpers for
  downloading Landsat, Sentinel-1, and Sentinel-2 data from free and
  public STAC servers providing **r**apid **s**atellite **i**magery,
- A **r**aster **s**tack **i**ntegration method for combining multiple
  rasters containing distinct data sets into a single raster stack.

The functions in rsi are designed around letting you use the tools
you’re familiar with to process raster data using compute that you
control – whether that means grabbing imagery with your laptop to add
some context to a map, or grabbing tranches of data to a virtual server
hosted near your data provider for lightning fast downloads. The outputs
from rsi functions are standard objects – usually the file paths of
raster files saved to your hard drive – meaning it’s easy to incorporate
rsi into broader spatial data processing workflows.

## Installation

You can install rsi via:

``` r
install.packages("rsi")
```

You can install the development version of rsi from
[GitHub](https://github.com/Permian-Global-Research/rsi) using
[pak](https://pak.r-lib.org/):

``` r
# install.packages("pak")
pak::pak("Permian-Global-Research/rsi")
```

## Example

The `spectral_indices()` function provides a tibble with data from the
[Awesome Spectral Indices
project](https://github.com/awesome-spectral-indices/awesome-spectral-indices):

``` r
library(rsi)

spectral_indices()
#> # A tibble: 245 × 9
#>    application_domain bands     contributor   date_of_addition formula long_name
#>    <chr>              <list>    <chr>         <chr>            <chr>   <chr>    
#>  1 vegetation         <chr [2]> https://gith… 2021-11-17       (N - 0… Aerosol …
#>  2 vegetation         <chr [2]> https://gith… 2021-11-17       (N - 0… Aerosol …
#>  3 water              <chr [6]> https://gith… 2022-09-22       (B + G… Augmente…
#>  4 vegetation         <chr [2]> https://gith… 2021-09-20       (1 / G… Anthocya…
#>  5 vegetation         <chr [3]> https://gith… 2022-04-08       N * ((… Anthocya…
#>  6 vegetation         <chr [4]> https://gith… 2021-05-11       (N - (… Atmosphe…
#>  7 vegetation         <chr [4]> https://gith… 2021-05-14       sla * … Adjusted…
#>  8 vegetation         <chr [2]> https://gith… 2022-04-08       (N * (… Advanced…
#>  9 water              <chr [4]> https://gith… 2021-09-18       4.0 * … Automate…
#> 10 water              <chr [5]> https://gith… 2021-09-18       B + 2.… Automate…
#> # ℹ 235 more rows
#> # ℹ 3 more variables: platforms <list>, reference <chr>, short_name <chr>
```

The first time `spectral_indices()` is called it will download the most
up-to-date version of the spectral indices JSON file, and then write the
resulting table to a cache file in `tools::R_user_dir("rsi")`. After
that, `spectral_indices()` will only download a new file if the cache is
older than 1 day, or if the `update_cache` argument is `TRUE`, in order
to provide the most up-to-date data as quickly as possible. If offline,
`spectral_indices()` will always fall back to the cache or, if no cache
file exists, a (possibly out-of-date) data file included in rsi itself.

Separately, the `get_stac_data()` function provides a generic interface
for downloading composite images from any accessible STAC catalog. For
instance, we could download a cloud-masked composite of Landsat’s
visible layers using `get_stac_data()` and a few helper functions from
rsi:

``` r
aoi <- sf::st_point(c(-74.912131, 44.080410))
aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 1000)

landsat_image <- get_stac_data(
  aoi,
  start_date = "2022-06-01",
  end_date = "2022-06-30",
  pixel_x_size = 30,
  pixel_y_size = 30,
  asset_names = c("red", "blue", "green"),
  stac_source = "https://planetarycomputer.microsoft.com/api/stac/v1/",
  collection = "landsat-c2-l2",
  mask_band = "qa_pixel",
  mask_function = landsat_mask_function,
  output_filename = tempfile(fileext = ".tif"),
  item_filter_function = landsat_platform_filter,
  platforms = c("landsat-9", "landsat-8")
)

terra::plot(terra::rast(landsat_image))
```

<img src="man/figures/README-unnamed-chunk-3-1.png" width="100%" />

For common data sets, rsi also provides helper functions which provide
most of these arguments for you. For instance, that `get_stac_data()`
call could be as simple as:

``` r
landsat_image <- get_landsat_imagery(
  aoi,
  start_date = "2022-06-01",
  end_date = "2022-08-30",
  output_filename = tempfile(fileext = ".tif")
)
terra::plot(terra::rast(landsat_image))
```

<img src="man/figures/README-unnamed-chunk-4-1.png" width="100%" />

Note that we’ve been plotting each band individually so far by calling
`terra::plot()`. We could also use `terra::plotRGB()` (after
`terra::stretch()`ing the band values) to see what this mosaic of images
would look like to the human eye:

``` r
landsat_image |>
  terra::rast(lyrs = c("R", "G", "B")) |>
  terra::stretch() |>
  terra::plotRGB()
```

<img src="man/figures/README-unnamed-chunk-5-1.png" width="100%" />

By default, these functions download data from Microsoft’s Planetary
Computer API, using a number of configuration options set in
`rsi_band_mapping` objects provided by the package. You can see these
default configuration options by printing the band mapping objects, and
can adjust them through arguments to any `get_*` function in the
package.

``` r
landsat_band_mapping$planetary_computer_v1
#> An rsi band mapping object with attributes:
#> names mask_band mask_function stac_source collection_name query_function download_function sign_function class
#> 
#> coastal    blue   green     red   nir08  swir16  swir22    lwir  lwir11 
#>     "A"     "B"     "G"     "R"     "N"    "S1"    "S2"     "T"    "T1"
```

We can put these pieces together and calculate as many spectral indices
as we can based on our downloaded Landsat imagery. The
`calculate_indices()` function, well, calculates indices, using subsets
of our `spectral_indices()` data frame:

``` r
available_indices <- filter_bands(
  bands = names(terra::rast(landsat_image))
)

indices <- calculate_indices(
  landsat_image,
  available_indices,
  output_filename = tempfile(fileext = ".tif")
)

# Plot the first handful of spatial indices
terra::plot(terra::rast(indices))
```

<img src="man/figures/README-unnamed-chunk-7-1.png" width="100%" />

And last but not least, rsi includes a utility for efficiently combining
rasters containing different data about the same location into a
[VRT](https://gdal.org/en/latest/drivers/raster/vrt.html), which allows
programs like GDAL to treat these separate data sources as a single
file. For instance, we can combine our Landsat imagery with the derived
indices:

``` r
raster_stack <- stack_rasters(
  c(landsat_image, indices),
  tempfile(fileext = ".vrt")
)

# The first few panels are now Landsat measurements, not indices:
terra::plot(terra::rast(raster_stack))
```

<img src="man/figures/README-unnamed-chunk-8-1.png" width="100%" />

This can be extremely useful as a way to create predictor bricks and
other multi-band rasters from various data sources.

## Contributing

We love contributions! See our [contribution
guide](https://github.com/Permian-Global-Research/rsi/blob/main/.github/CONTRIBUTING.md)
for pointers on how to make your contribution as easy to accept as
possible – in particular, consider [opening an
issue](https://github.com/Permian-Global-Research/rsi/issues/new/choose)
with a [minimal reprex](https://tidyverse.org/help/#reprex) to make sure
that we understand what your changes are meant to do.

## License

Copyright 2023 Permian Global Research, Limited.

Licensed under the Apache License, Version 2.0 (the “License”); you may
not use this file except in compliance with the License.

You may obtain a copy of the License at:

<https://www.apache.org/licenses/LICENSE-2.0>

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an “AS IS” BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

#' Download specific assets from a set of STAC items
#'
#' @param items A `StacItemCollection` object, as returned by [rsi_query_api()].
#' @param aoi Either an sf(c) object outlining the area of interest to get
#' imagery for, or a `bbox` image containing the bounding box of your AOI.
#' @param merge Logical: for each asset, should data from multiple items be
#' merged into a single downloaded file? If `TRUE`, this returns a single file
#' for each asset, which has been merged via gdalwarp. No resampling or
#' compositing is performed, but rather each pixel uses the last data
#' downloaded. This is fast, but precludes per-item masking and compositing.
#' If `FALSE`, each asset from each item is saved as a separate file.
#' @inheritParams get_stac_data
#'
#' @returns A data frame  where columns correspond to distinct assets, rows
#' correspond to distinct items, and cells contain file paths to the downloaded
#' data.
#'
#' @examplesIf interactive()
#' aoi <- sf::st_point(c(-74.912131, 44.080410))
#' aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
#' aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)
#'
#' landsat_image <- get_landsat_imagery(
#'   aoi,
#'   start_date = "2022-06-01",
#'   end_date = "2022-08-30",
#'   download_function = rsi_download_rasters
#' )
#'
#' @export
rsi_download_rasters <- function(
  items,
  aoi,
  asset_names,
  sign_function = NULL,
  merge = FALSE,
  gdalwarp_options = c(
    "-r",
    "bilinear",
    "-multi",
    "-overwrite",
    "-co",
    "COMPRESS=DEFLATE",
    "-co",
    "PREDICTOR=2",
    "-co",
    "NUM_THREADS=ALL_CPUS"
  ),
  gdal_config_options = c(
    VSI_CACHE = "TRUE",
    GDAL_CACHEMAX = "30%",
    VSI_CACHE_SIZE = "10000000",
    GDAL_HTTP_MULTIPLEX = "YES",
    GDAL_INGESTED_BYTES_AT_OPEN = "32000",
    GDAL_DISABLE_READDIR_ON_OPEN = "EMPTY_DIR",
    GDAL_HTTP_VERSION = "2",
    GDAL_HTTP_MERGE_CONSECUTIVE_RANGES = "YES",
    GDAL_NUM_THREADS = "ALL_CPUS",
    GDAL_HTTP_USERAGENT = "rsi (https://permian-global-research.github.io/rsi/)"
  ),
  ...
) {
  if (!inherits(aoi, "bbox")) aoi <- sf::st_bbox(aoi)

  check_type_and_length(
    merge = logical(1)
  )

  n_tiles_out <- ifelse(merge, 1L, length(items$features))
  p <- build_progressr(length(names(asset_names)) * n_tiles_out)

  download_locations <- data.frame(
    matrix(
      data = replicate(
        length(asset_names) * n_tiles_out,
        tempfile(fileext = ".tif")
      ),
      ncol = length(asset_names),
      nrow = n_tiles_out
    )
  )
  names(download_locations) <- names(asset_names)

  if (merge) {
    gdalwarp_options <- set_gdalwarp_extent(gdalwarp_options, aoi, NULL)
  }

  # Which loop gets parallelized is determined based on which has more steps,
  # working from the assumption that all downloads take about as long
  #
  # so if we aren't merging, or if there's more assets than tiles, we'll
  # parallelize the outside loop that walks over assets (which, unless the user
  # has set up wild nested futures, turns the inside one into a serial process)
  asset_iterator <- ifelse(
    merge || (n_tiles_out < ncol(download_locations)),
    function(...) future.apply::future_lapply(..., future.seed = TRUE),
    lapply
  )

  current_options <- gdalwarp_options

  asset_iterator(
    names(download_locations),
    function(asset) {
      feature_iter <- seq_len(length(items$features))
      if (length(download_locations[[asset]]) == 1) {
        feature_iter <- list(feature_iter)
      }

      future.apply::future_mapply(
        function(which_item, dl_location) {
          p(glue::glue("Downloading {asset}"))
          signed_items <- maybe_sign_items(items, sign_function)
          url <- rstac::assets_url(signed_items, asset)[which_item]

          if (!merge) {
            item_bbox <- items$features[[which_item]]$bbox
            current_options <- set_gdalwarp_extent(
              gdalwarp_options,
              aoi,
              item_bbox
            )
          }

          tryCatch(
            {
              sf::gdal_utils(
                "warp",
                paste0("/vsicurl/", url),
                dl_location,
                options = current_options,
                quiet = TRUE,
                config_options = gdal_config_options
              )
            },
            error = function(e) {
              # stop if failure occurs when merging.
              if (merge) {
                rlang::abort(
                  glue::glue(
                    "GDAL warp failed when attempting to merge ",
                    "{length(unlist(feature_iter))} items"
                  ),
                  class = "rsi_download_warp_error",
                  parent = e
                )
              }

              expr1 <- items$features[[which_item]]$id %||% "UNKNOWN"
              expr2 <- items$features[[which_item]]$properties$datetime %||%
                "UNKNOWN"

              err_msg <- glue::glue(
                "Failed to download {expr1} from {expr2}" # nolint
              )

              # stop if failure occurs when there is only one item to download.
              if (length(feature_iter) == 1) {
                rlang::abort(
                  err_msg,
                  class = "rsi_download_warp_error",
                  parent = e
                )
              }

              # warn if failure occurs when multiple items are being downloaded.
              rlang::warn(err_msg, parent = e)
              download_locations[which_item, ] <<- NA
            }
          )
        },
        which_item = feature_iter,
        dl_location = download_locations[[asset]],
        future.seed = TRUE
      )
    }
  )
  out <- stats::na.omit(download_locations)
  na_attr <- stats::na.action(out)
  out <- as.data.frame(as.list(out))
  attr(out, "na.action") <- na_attr
  out
}

maybe_sign_items <- function(items, sign_function) {
  if (!is.null(sign_function)) {
    items <- sign_function(items)
  }
  items
}

set_gdalwarp_extent <- function(gdalwarp_options, aoi_bbox, item_bbox = NULL) {
  if (!("-te" %in% gdalwarp_options)) {
    if (!is.null(item_bbox)) {
      class(item_bbox) <- "bbox"
      item_bbox <- sf::st_as_sfc(item_bbox)
      item_bbox <- sf::st_set_crs(item_bbox, 4326)
      item_bbox <- sf::st_transform(item_bbox, sf::st_crs(aoi_bbox))
      item_bbox <- sf::st_bbox(item_bbox)

      aoi_bbox <- c(
        xmin = max(aoi_bbox[[1]], item_bbox[[1]]),
        ymin = max(aoi_bbox[[2]], item_bbox[[2]]),
        xmax = min(aoi_bbox[[3]], item_bbox[[3]]),
        ymax = min(aoi_bbox[[4]], item_bbox[[4]])
      )

      aoi_bbox <- c(
        xmin = min(aoi_bbox[["xmin"]], aoi_bbox[["xmax"]]),
        ymin = min(aoi_bbox[["ymin"]], aoi_bbox[["ymax"]]),
        xmax = max(aoi_bbox[["xmin"]], aoi_bbox[["xmax"]]),
        ymax = max(aoi_bbox[["ymin"]], aoi_bbox[["ymax"]])
      )
    }

    gdalwarp_options <- c(gdalwarp_options, "-te", aoi_bbox)
  }
  gdalwarp_options
}

#' @export
`[.rsi_band_mapping` <- function(x, i) {
  out <- x
  attributes(out) <- NULL
  names(out) <- names(x)
  out <- out[i]
  out_names <- names(out)
  mostattributes(out) <- attributes(x)
  names(out) <- out_names
  out
}

#' @export
`[[.rsi_band_mapping` <- function(x, i) {
  out <- x
  attributes(out) <- NULL
  names(out) <- names(x)
  out <- out[i]
  out_names <- names(out)
  mostattributes(out) <- attributes(x)
  names(out) <- out_names
  out
}

#' @export
c.rsi_band_mapping <- function(...) {
  dots <- list(...)
  x <- dots[[1]]
  dots[[1]] <- as.vector(x)
  names(dots[[1]]) <- names(x)
  out <- do.call(c, dots)
  out_names <- names(out)
  mostattributes(out) <- attributes(x)
  names(out) <- out_names
  not_named <- which(names(out) == "")
  names(out)[not_named] <- out[not_named]
  class(out) <- class(x)
  out
}

#' @export
print.rsi_band_mapping <- function(x, ...) {
  cat("An rsi band mapping object with attributes:\n")
  cat(names(attributes(x)), sep = " ")
  cat("\n\n")
  x_names <- names(x)
  attributes(x) <- NULL
  names(x) <- x_names
  print(x)
}

build_progressr <- function(n) {
  if (rlang::is_installed("progressr")) {
    progressr::progressor(n, on_exit = FALSE)
  } else {
    function(...) NULL # nocov
  }
}

#' Default options for GDAL
#'
#' These functions provide useful default options for GDAL functions,
#' making downloading and warping (hopefully!) more efficient for
#' most use cases.
#'
#' @returns A vector of options for GDAL commands.
#'
#' @name rsi_gdal_options
#' @export
rsi_gdal_config_options <- function() {
  c(
    VSI_CACHE = "TRUE",
    GDAL_CACHEMAX = "30%",
    VSI_CACHE_SIZE = "10000000",
    GDAL_HTTP_MULTIPLEX = "YES",
    GDAL_INGESTED_BYTES_AT_OPEN = "32000",
    GDAL_DISABLE_READDIR_ON_OPEN = "EMPTY_DIR",
    GDAL_HTTP_VERSION = "2",
    GDAL_HTTP_MERGE_CONSECUTIVE_RANGES = "YES",
    GDAL_NUM_THREADS = "ALL_CPUS",
    GDAL_HTTP_USERAGENT = "rsi (https://permian-global-research.github.io/rsi/)"
  )
}

#' @rdname rsi_gdal_options
#' @export
rsi_gdalwarp_options <- function() {
  c(
    "-r",
    "bilinear",
    "-multi",
    "-overwrite",
    "-co",
    "COMPRESS=DEFLATE",
    "-co",
    "PREDICTOR=2",
    "-co",
    "NUM_THREADS=ALL_CPUS"
  )
}

#' Query a STAC API using a specific spatiotemporal area of interest
#'
#' This function is the default method used to retrieve lists of items to
#' download for all the collections and endpoints supported by rsi. It will
#' likely work for any other STAC APIs of interest.
#'
#' You can pass your own query functions to `get_stac_data()` and its variants.
#' This is the best way to perform more complex queries, for instance if you
#' need to provide authentication to get the list of items (not just the assets)
#' available for your AOI, or to perform cloud filtering prior to downloading
#' assets.
#'
#' @param bbox A `bbox` or `sfc` object, from the sf package, representing the
#' spatial bounding box of your area of interest. This must be in EPSG:4326
#' coordinates (and, if this function is called from within `get_stac_data()`,
#' it will be) or else it will be automatically reprojected.
#' @inheritParams get_stac_data
#' @param start_date,end_date Character strings of length 1 representing the
#' boundaries of your temporal range of interest, in RFC-3339 format. Set either
#' argument to `..` to use an open interval; set `start_date` to `NULL` to not
#' pass a temporal range of interest (which may cause errors with some APIs). If
#' this function is called from within `get_stac_data()`, the inputs to
#' `start_date` and `end_date` will have already been processed to try and force
#' RFC-3339 compliance.
#' @param ... Ignored by this function. Arguments passed to `get_stac_data()`
#' via `...` will be available (unchanged) in this function
#'
#' @returns  A StacItemCollection object.
#'
#' @examplesIf interactive()
#' aoi <- sf::st_point(c(-74.912131, 44.080410))
#' aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
#' aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)
#'
#' landsat_image <- get_landsat_imagery(
#'   aoi,
#'   start_date = "2022-06-01",
#'   end_date = "2022-08-30",
#'   query_function = rsi_query_api
#' )
#'
#' @export
rsi_query_api <- function(
  bbox,
  stac_source,
  collection,
  start_date,
  end_date,
  limit,
  ...
) {
  if (!is.null(start_date)) {
    datetime <- paste0(start_date, "/", end_date)
  } else {
    datetime <- NULL
  }

  if (!(inherits(bbox, "bbox") | inherits(bbox, "sfc"))) {
    rlang::abort(
      "`bbox` must be either an sfc or a bbox object from the sf package.",
      class = "rsi_bbox_wrong_class"
    )
  }

  if (!(sf::st_crs(bbox) == sf::st_crs("EPSG:4326"))) {
    rlang::warn(
      "Reprojecting `bbox` to EPSG:4326.",
      class = "rsi_reprojecting_bbox"
    )
    if (inherits(bbox, "bbox")) {
      bbox <- sf::st_as_sfc(bbox)
    }
    bbox <- sf::st_transform(bbox, 4326)
  }

  if (inherits(bbox, "sfc")) bbox <- sf::st_bbox(bbox)

  items <- rstac::stac_search(
    rstac::stac(stac_source),
    collections = collection,
    bbox = c(
      bbox["xmin"],
      bbox["ymin"],
      bbox["xmax"],
      bbox["ymax"]
    ),
    datetime = datetime,
    limit = limit
  )

  items <- rstac::items_fetch(
    rstac::get_request(items, rsi_user_agent),
    rsi_user_agent
  )

  items
}

#' Sign STAC items retrieved from the Planetary Computer
#'
#' @param items A STACItemCollection, as returned by `rsi_query_api`.
#' @param subscription_key Optionally, a subscription key associated with your
#' Planetary Computer account. At the time of writing, this is required for
#' downloading Sentinel 1 RTC products, as well as NAIP imagery. This key will
#' be automatically used if the environment variable `rsi_pc_key` is set.
#'
#' @returns A STACItemCollection object with signed assets url.
#'
#' @examplesIf interactive()
#' aoi <- sf::st_point(c(-74.912131, 44.080410))
#' aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
#' aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)
#'
#' landsat_image <- get_landsat_imagery(
#'   aoi,
#'   start_date = "2022-06-01",
#'   end_date = "2022-08-30",
#'   sign_function = sign_planetary_computer
#' )
#'
#' @export
sign_planetary_computer <- function(
  items,
  subscription_key = Sys.getenv("rsi_pc_key")
) {
  # check for the variable used by sits if the rsi one isn't set
  if (subscription_key == "") subscription_key <- Sys.getenv("MPC_TOKEN")
  if (subscription_key == "") {
    rstac::items_sign(items, rstac::sign_planetary_computer(rsi_user_agent))
  } else {
    rstac::items_sign(
      items,
      rstac::sign_planetary_computer(
        rsi_user_agent,
        headers = c("Ocp-Apim-Subscription-Key" = subscription_key)
      )
    )
  }
}

test_that("non-4326 CRS warns", {
  skip_on_cran()
  skip_if_offline()

  nc <- sf::read_sf(
    system.file("shape/nc.shp", package = "sf")
  )
  expect_warning(
    rsi_query_api(
      sf::st_as_sfc(sf::st_bbox(nc)),
      stac_source = "https://planetarycomputer.microsoft.com/api/stac/v1/",
      collection = "landsat-c2-l2",
      start_date = "2023-08-01",
      end_date = "2023-09-01",
      limit = 10
    ),
    class = "rsi_reprojecting_bbox"
  )
})

test_that("unaccepted bbox objects error well", {
  expect_error(
    rsi_query_api(
      "not a bbox",
      stac_source = "https://planetarycomputer.microsoft.com/api/stac/v1/",
      collection = "landsat-c2-l2",
      start_date = "2023-08-01",
      end_date = "2023-09-01",
      limit = 10
    ),
    class = "rsi_bbox_wrong_class"
  )
})

#' Get the URL to download spectral indices from
#'
#' @returns A URL to download indices from.
#'
#' @examples
#' spectral_indices_url()
#'
#' @export
spectral_indices_url <- function() {
  if (!is.null(getOption("rsi_url"))) {
    return(getOption("rsi_url"))
  }
  if (!(Sys.getenv("rsi_url") == "")) {
    return(Sys.getenv("rsi_url"))
  }
  "https://raw.githubusercontent.com/awesome-spectral-indices/awesome-spectral-indices/main/output/spectral-indices-dict.json"
}

#' Get a data frame of spectral indices
#'
#' This function returns a data frame of spectral indices, from the
#' `awesome-spectral-indices` repository.
#'
#' @inheritParams rlang::args_dots_empty
#' @param url The URL to download spectral indices from. If the option `rsi_url`
#' is set, that value will be used; otherwise, if the environment variable
#' `rsi_url` is set, that value will be used; otherwise, the list at
#' https://github.com/awesome-spectral-indices/awesome-spectral-indices will
#' be used.
#' @param download_indices Logical: should this function download indices? If
#' `NULL`, this function will only download indices if the cache will be
#' updated. If `TRUE`, this function will attempt to download indices no matter
#' what. If `FALSE`, either cached or package indices will be used.
#' @param update_cache Logical: should cached indices be updated? If `NULL`,
#' cached values will be updated if the cache is older than a day. If `TRUE`,
#' the cache will be updated, if `FALSE` it will not.
#'
#' @examples
#' spectral_indices()
#'
#' @returns A [tibble::tibble] with nine columns, containing information about spectral indices.
#'
#' @source [https://github.com/awesome-spectral-indices/awesome-spectral-indices](https://github.com/awesome-spectral-indices/awesome-spectral-indices)
#'
#' @export
spectral_indices <- function(
  ...,
  url = spectral_indices_url(),
  download_indices = NULL,
  update_cache = NULL
) {
  rlang::check_dots_empty()
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")

  if (isFALSE(download_indices)) {
    if (isTRUE(update_cache)) {
      rlang::abort(
        "Cannot update the cache if not downloading indices.",
        class = "rsi_cache_download_conflict"
      )
    }

    update_cache <- FALSE
  }

  if (is.null(update_cache)) {
    if (file.exists(indices_path)) {
      update_cache <- (Sys.time() - file.info(indices_path)[["mtime"]]) > 86400
    } else {
      update_cache <- TRUE
    }
  }

  if (is.null(download_indices)) download_indices <- update_cache

  if (update_cache && isTRUE(download_indices)) {
    tryCatch(
      update_cached_indices(url),
      error = function(e) {
        # nocov start
        rlang::warn(
          c(
            "Failed to update the cache of indices.",
            i = "Returning (likely outdated) cached data instead."
          ),
          class = "rsi_failed_cache_update"
        )
        spectral_indices_internal
      } # nocov end
    )
  }

  if (!isTRUE(download_indices) && file.exists(indices_path)) {
    tibble::as_tibble(readRDS(indices_path))
  } else if (isTRUE(download_indices)) {
    tryCatch(
      download_web_indices(url),
      error = function(e) {
        # nocov start
        rlang::warn(
          c(
            "Failed to download new indices.",
            i = "Returning (likely outdated) package data instead."
          ),
          class = "rsi_failed_download"
        )
        spectral_indices_internal
      }
    ) # nocov end
  } else {
    rlang::warn(
      c(
        "No cache file present and `download_indices` set to `FALSE`.",
        i = "Returning (likely outdated) package data instead."
      ),
      class = "rsi_failed_download"
    )
    spectral_indices_internal
  }
}

download_web_indices <- function(url = spectral_indices_url()) {
  spectral_indices <- lapply(
    jsonlite::read_json(url)[[1]],
    function(index) {
      for (col in names(index)) {
        if (length(index[[col]]) > 1) index[[col]] <- list(unlist(index[[col]]))
      }
      tibble::as_tibble(index)
    }
  )

  spectral_indices <- do.call(rbind, spectral_indices)

  spectral_indices$formula <- gsub(" \\*\\* ", "\\^", spectral_indices$formula)
  spectral_indices$formula <- gsub("\\*\\*", "\\^", spectral_indices$formula)
  spectral_indices
}

update_cached_indices <- function(url = spectral_indices_url()) {
  # nocov start
  if (!dir.exists(tools::R_user_dir("rsi"))) {
    dir.create(tools::R_user_dir("rsi"), recursive = TRUE)
  }
  # nocov end
  indices_path <- file.path(
    tools::R_user_dir("rsi"),
    "indices.rda"
  )
  saveRDS(
    download_web_indices(url),
    indices_path
  )
}

test_that("spectral_indices() works", {
  # Loading indices three ways
  # First, even if we have no internet and no cache (like on CRAN)
  # then we should still load indices from package data, if needed,
  # with no warnings other than the package's warning:
  expect_no_warning(
    suppressWarnings(
      spectral_indices(update_cache = FALSE),
      classes = "rsi_failed_download"
    )
  )

  # github actions seems to not like reading from a github URL
  skip_on_ci()

  # Second, if we're online, we should be able to download the current indices:
  skip_if_offline()
  skip_on_cran()
  expect_no_warning(spectral_indices(
    update_cache = FALSE,
    download_indices = TRUE
  ))

  # Third, if we're online and can write to the cache folder,
  # we should be able to update our cache:
  skip_on_cran()
  expect_no_warning(spectral_indices())
  expect_no_warning(spectral_indices(update_cache = TRUE))
})

test_that("spectral_indices_url() respects options", {
  skip_if_not_installed("withr")
  expect_identical(
    withr::with_options(
      list("rsi_url" = "example"),
      spectral_indices_url()
    ),
    "example"
  )
})

test_that("spectral_indices_url() respects environment variables", {
  skip_if_not_installed("withr")
  expect_identical(
    withr::with_envvar(
      list("rsi_url" = "example"),
      spectral_indices_url()
    ),
    "example"
  )
})

test_that("no cache", {
  skip_if_offline()
  skip_on_cran()
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")
  if (file.exists(indices_path)) file.remove(indices_path)
  expect_no_warning(spectral_indices())
})

test_that("no cache, update false", {
  skip_if_offline()
  skip_on_cran()
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")
  if (file.exists(indices_path)) file.remove(indices_path)
  expect_snapshot(suppressWarnings(spectral_indices(update_cache = FALSE)))
})

test_that("no cache, download false", {
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")
  if (file.exists(indices_path)) file.remove(indices_path)
  expect_snapshot(spectral_indices(download_indices = FALSE))
})

test_that("no cache, download and update false", {
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")
  if (file.exists(indices_path)) file.remove(indices_path)
  expect_snapshot(spectral_indices(
    download_indices = FALSE,
    update_cache = FALSE
  ))
})

test_that("no cache, update true", {
  skip_if_offline()
  skip_on_cran()
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")
  if (file.exists(indices_path)) file.remove(indices_path)
  expect_no_warning(spectral_indices(update_cache = TRUE))
})

test_that("no cache, download true", {
  skip_if_offline()
  skip_on_cran()
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")
  if (file.exists(indices_path)) file.remove(indices_path)
  expect_no_warning(spectral_indices(download_indices = TRUE))
})

test_that("no cache, download and update true", {
  skip_if_offline()
  skip_on_cran()
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")
  if (file.exists(indices_path)) file.remove(indices_path)
  expect_no_warning(spectral_indices(
    download_indices = TRUE,
    update_cache = TRUE
  ))
})

test_that("new cache", {
  skip_if_offline()
  skip_on_cran()
  invisible(spectral_indices(update_cache = TRUE))
  expect_no_warning(spectral_indices())
})

test_that("new cache, update false", {
  skip_if_offline()
  skip_on_cran()
  invisible(spectral_indices(update_cache = TRUE))
  expect_no_warning(spectral_indices(update_cache = FALSE))
})

test_that("new cache, download false", {
  skip_if_offline()
  skip_on_cran()
  invisible(spectral_indices(update_cache = TRUE))
  expect_no_warning(spectral_indices(download_indices = FALSE))
})

test_that("new cache, download and update false", {
  skip_if_offline()
  skip_on_cran()
  invisible(spectral_indices(update_cache = TRUE))
  expect_no_warning(spectral_indices(
    download_indices = FALSE,
    update_cache = FALSE
  ))
})

test_that("new cache, update true", {
  skip_if_offline()
  skip_on_cran()
  invisible(spectral_indices(update_cache = TRUE))
  expect_no_warning(spectral_indices(update_cache = TRUE))
})

test_that("new cache, download true", {
  skip_if_offline()
  skip_on_cran()
  invisible(spectral_indices(update_cache = TRUE))
  expect_no_warning(spectral_indices(download_indices = TRUE))
})

test_that("new cache, download and update true", {
  skip_if_offline()
  skip_on_cran()
  invisible(spectral_indices(update_cache = TRUE))
  expect_no_warning(spectral_indices(
    download_indices = TRUE,
    update_cache = TRUE
  ))
})

test_that("old cache", {
  skip_if_offline()
  skip_on_cran()
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")
  invisible(spectral_indices(update_cache = TRUE))
  Sys.setFileTime(indices_path, "1970-01-01")
  expect_no_warning(spectral_indices())
})

test_that("old cache, update false", {
  skip_if_offline()
  skip_on_cran()
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")
  invisible(spectral_indices(update_cache = TRUE))
  Sys.setFileTime(indices_path, "1970-01-01")
  expect_no_warning(spectral_indices(update_cache = FALSE))
})

test_that("old cache, download false", {
  skip_if_offline()
  skip_on_cran()
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")
  invisible(spectral_indices(update_cache = TRUE))
  Sys.setFileTime(indices_path, "1970-01-01")
  expect_no_warning(spectral_indices(download_indices = FALSE))
})

test_that("old cache, download and update false", {
  skip_if_offline()
  skip_on_cran()
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")
  invisible(spectral_indices(update_cache = TRUE))
  Sys.setFileTime(indices_path, "1970-01-01")
  expect_no_warning(spectral_indices(
    download_indices = FALSE,
    update_cache = FALSE
  ))
})

test_that("old cache, update true", {
  skip_if_offline()
  skip_on_cran()
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")
  invisible(spectral_indices(update_cache = TRUE))
  Sys.setFileTime(indices_path, "1970-01-01")
  expect_no_warning(spectral_indices(update_cache = TRUE))
})

test_that("old cache, download true", {
  skip_if_offline()
  skip_on_cran()
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")
  invisible(spectral_indices(update_cache = TRUE))
  Sys.setFileTime(indices_path, "1970-01-01")
  expect_no_warning(spectral_indices(download_indices = TRUE))
})

test_that("old cache, download and update true", {
  skip_if_offline()
  skip_on_cran()
  indices_path <- file.path(tools::R_user_dir("rsi"), "indices.rda")
  invisible(spectral_indices(update_cache = TRUE))
  Sys.setFileTime(indices_path, "1970-01-01")
  expect_no_warning(spectral_indices(
    download_indices = TRUE,
    update_cache = TRUE
  ))
})

test_that("download false and update true", {
  expect_snapshot(
    spectral_indices(download_indices = FALSE, update_cache = TRUE),
    error = TRUE
  )
})

test_that("Index calculation is stable", {
  skip_if_not_installed("terra")
  # covr can't instrument the local block properly
  skip_if(nzchar(Sys.getenv("is_covr")))
  skip_on_cran()
  index_out <- tempfile(fileext = ".tif")
  expect_no_error(
    out <- calculate_indices(
      system.file("rasters/example_sentinel1.tif", package = "rsi"),
      suppressWarnings(filter_platforms(
        spectral_indices(download_indices = FALSE, update_cache = FALSE),
        platforms = "Sentinel-1 (Dual Polarisation VV-VH)"
      )),
      index_out,
      names_suffix = "sentinel1"
    )
  )

  expect_no_error(terra::rast(out))

  expect_equal(
    terra::values(terra::rast(out, lyrs = "DPDD_sentinel1")),
    terra::values(terra::rast(system.file("rasters/dpdd.tif", package = "rsi")))
  )
})

test_that("Extra objects can be passed to calculate_indices()", {
  skip_if_not_installed("terra")
  # covr can't instrument the local block properly
  skip_if(nzchar(Sys.getenv("is_covr")))
  skip_on_cran()
  index_out <- tempfile(fileext = ".tif")

  idx <- suppressWarnings(filter_platforms(
    spectral_indices(download_indices = FALSE, update_cache = FALSE),
    platforms = "Sentinel-1 (Dual Polarisation VV-VH)"
  ))[1, ]
  idx$formula <- "pmax(VH, 1000)"

  expect_warning(
    out <- calculate_indices(
      system.file("rasters/example_sentinel1.tif", package = "rsi"),
      idx,
      index_out,
      names_suffix = "sentinel1",
      extra_objects = list(`pmax` = pmax)
    ),
    class = "rsi_extra_objects"
  )

  expect_true(
    all(terra::values(terra::rast(out)) == 1000)
  )
})

test_that("Index calculations fail when missing a column", {
  skip_on_cran()
  index_out <- tempfile(fileext = ".tif")
  expect_error(
    calculate_indices(
      system.file("rasters/example_sentinel1.tif", package = "rsi"),
      suppressWarnings(filter_platforms(
        spectral_indices(download_indices = FALSE, update_cache = FALSE),
        platforms = "Sentinel-1 (Dual Polarisation VV-VH)"
      ))["formula"],
      index_out
    ),
    class = "rsi_missing_column"
  )
})

test_that("Index calculations fail when missing bands", {
  skip_on_cran()
  index_out <- tempfile(fileext = ".tif")
  expect_error(
    calculate_indices(
      system.file("rasters/example_sentinel1.tif", package = "rsi"),
      suppressWarnings(filter_platforms(
        spectral_indices(download_indices = FALSE, update_cache = FALSE),
        platforms = "Landsat-OLI"
      )),
      index_out
    ),
    class = "rsi_missing_indices"
  )
})

test_that("Index calculations stop obvious security issues", {
  skip_on_cran()
  example_indices <- suppressWarnings(filter_platforms(
    spectral_indices(download_indices = FALSE, update_cache = FALSE),
    platforms = "Sentinel-1 (Dual Polarisation VV-VH)"
  ))[1, ]
  example_indices$formula <- 'system("echo something bad")'
  expect_error(calculate_indices(
    system.file("rasters/example_sentinel1.tif", package = "rsi"),
    example_indices,
    tempfile(fileext = ".tif")
  ))
})

#' Retrieve raster data from STAC endpoints
#'
#' These functions retrieve raster data from STAC endpoints and optionally
#' create composite data sets from multiple files.
#' `get_stac_data()` is a generic function which should be able to download
#' raster data from a variety of data sources, while the other helper functions
#' have useful defaults for downloading common data sets from standard
#' STAC sources.
#'
#' @section Usage Tips:
#' It's often useful to buffer your `aoi` object slightly, on the order of 1-2
#' cell widths, in order to ensure that data is downloaded for your entire AOI
#' even after accounting for any reprojection needed to compare your AOI to
#' the data on the STAC server.
#'
#' These functions allow for parallelizing downloads via [future::plan()], and
#' for user-controlled progress updates via [progressr::handlers()]. If
#' there are fewer images to download than `asset_names`, then this function
#' uses [lapply()] to iterate through images and [future.apply::future_mapply()]
#' to iterate through downloading each asset. If there are more images than
#' assets, this function uses [future.apply::future_lapply()] to iterate through
#' images.
#'
#' @section Downloading from Planetary Computer:
#'
#' Certain data sets in Planetary Computer require
#' [providing a subscription key](https://planetarycomputer.microsoft.com/docs/concepts/sas/).
#' Even for non-protected data sets, providing a subscription key grants you
#' higher rate limits and faster downloads. As such, it's a good idea to
#' [request a Planetary Computer account](https://planetarycomputer.microsoft.com/account/request),
#' then [generate a subscription key](https://planetarycomputer.developer.azure-api.net/).
#' If you set the `rsi_pc_key` environment variable to your key (either primary
#' or secondary; there is no difference), rsi will automatically use
#' this key to sign all requests against Planetary Computer.
#'
#' There are currently some challenges with certain Landsat images in Planetary
#' Computer; please see
#' https://github.com/microsoft/PlanetaryComputer/discussions/101
#' for more information on these images and their current status. These files
#' may cause data downloads to fail.
#'
#' @section Compositing:
#'
#' This function can either download all data that intersects with your
#' spatiotemporal AOI as multiple files (if `composite_function = NULL`),
#' or can be used to rescale band values, apply a mask function, and create a
#' composite from the resulting files in a single function call. Each of these
#' steps can be skipped by passing `NULL` to the corresponding argument.
#'
#' Masks are applied to each downloaded asset separately. Rescaling is applied
#' to the final composite after images are combined.
#'
#' A number of the steps involved in creating composites -- rescaling band
#' values, running the mask function, masking images, and compositing images --
#' currently rely on the `terra` package for raster calculations. This means
#' creating larger composites, either in geographic or temporal dimension, may
#' cause errors. It can be a good idea to tile your `aoi` using
#' `sf::st_make_grid()` and iterate through the tiles to avoid these errors
#' (and to make it easier to interrupt and restart a download job).
#'
#' @section Rescaling:
#' If `rescale_bands` is `TRUE`, then this function is able to use the `scale`
#' and `offset` values in the `bands` field of the `raster` STAC extension.
#' This was implemented originally to work with the Landsat collections in the
#' Planetary Computer STAC catalogue, but hopefully will work automatically with
#' other data sources as well. Note that Sentinel-2 data typically doesn't use
#' this STAC extension, and so the returned data is typically not re-scaled;
#' divide the downloaded band values by 10000 to get reflectance values in the
#' expected values.
#'
#' @section Sentinel-1 Data:
#' The `get_sentinel1_imagery()` function is designed to download Sentinel-1 data
#' from the Microsoft Planetary Computer STAC API. Both the GRD and RTC
#' Sentinel-1 collections are supported. To download RTC data,
#' set `collection` to `sentinel-1-rtc`, and supply your subscription key
#' as an environment variable named `rsi_pc_key` (through, e.g., `Sys.setenv()`
#' or your `.Renviron` file).
#'
#' @section AlOS PALSAR Data:
#' The `get_alos_palsar_imagery()` function is designed to download ALOS PALSAR
#' annual mosaic data from the Microsoft Planetary Computer STAC API. Data are
#' returned as a digital number (which is appropriate for some applications
#' and indices). To convert to backscatter (decibels) use the following formula:
#' `10 * log10(dn) - 83.0` where dn is the radar band in digital number.
#'
#' @param aoi An sf(c) object outlining the area of interest to get imagery for.
#' Will be used to get the bounding box used for calculating metrics and the
#' output data's CRS.
#' @param start_date,end_date Character of length 1: The first and last date,
#' respectively, of imagery to include in metrics calculations. Should be in
#' YYYY-MM-DD format.
#' @param pixel_x_size,pixel_y_size Numeric of length 1: size of pixels in
#' x-direction (longitude / easting) and y-direction (latitude / northing).
#' @param asset_names The names of the assets to download. If this vector has
#' names, then the names of the vector are assumed to be the names of assets on
#' the STAC server, which will be renamed to the elements of the vector in the
#' final output.
#' @param stac_source Character of length 1: the STAC URL to download
#' imagery from.
#' @param collection Character of length 1: the STAC collection to download
#' images from.
#' @param query_function A function that takes the output from
#' [rstac::stac_search()] and executes the request. See
#' [rsi_query_api()] and the `query_function` slots of
#' [sentinel1_band_mapping], [sentinel2_band_mapping], and
#' [landsat_band_mapping].
#' @param download_function A function that takes the output from
#' `query_function` and downloads the assets attached to those items. See
#' [rsi_download_rasters()] for an example.
#' @param sign_function A function that takes the output from `query_function`
#' and signs the item URLs, if necessary.
#' @param ... Passed to `item_filter_function`.
#' @param rescale_bands Logical of length 1: If the STAC collection implements
#' the `raster` STAC extension, and that extension includes `scale` and `offset`
#' values, should this function attempt to automatically rescale the downloaded
#' data?
#' @param item_filter_function A function that takes the outputs of
#' `query_function` (usually a `STACItemCollection`) and `...` and returns a
#' filtered `STACItemCollection`. This is used, for example, to only download
#' images from specific Landsat platforms.
#' @param mask_band Character of length 1: The name of the asset in your
#' STAC source to use to mask the data. Set to `NULL` to not mask. See the
#' `mask_band` slots of [sentinel1_band_mapping], [sentinel2_band_mapping], and
#' [landsat_band_mapping].
#' @param mask_function A function that takes a raster and returns a boolean
#' raster, where `TRUE` pixels will be preserved and `FALSE` or `NA` pixels will
#' be masked out. See [sentinel2_mask_function()].
#' @param output_filename The filename to write the output raster to. If
#' `composite_function` is `NULL`, item datetimes will be appended to this
#' in order to create unique filenames. If items do not have datetimes, a
#' sequential ID will be appended instead.
#' @param composite_function Character of length 1: The name of a
#' function used to combine downloaded images into a single composite
#' (i.e., to aggregate pixel values from multiple images into a single value).
#' Options include "merge", which 'stamps' images on top of one another such that
#' the "last" value downloaded for a pixel -- which isn't guaranteed to be the most
#' recent one -- will be the only value used, or any of "sum", "mean", "median",
#' "min", or "max", which consider all values available at each pixel.
#' Set to `NULL` to not composite
#' (i.e., to rescale and save each individual file independently).
#' @inheritParams rstac::stac_search
#' @param gdalwarp_options Options passed to `gdalwarp` through the `options`
#' argument of [sf::gdal_utils()]. The same set of options are used for all
#' downloaded data and the final output images; this means that some common
#' options (for instance, `PREDICTOR=3`) may cause errors if bands are of
#' varying data types. The default values are provided by
#' [rsi_gdalwarp_options()].
#' @param gdal_config_options Options passed to `gdalwarp` through the
#' `config_options` argument of [sf::gdal_utils()]. The default values are
#' provided by [rsi_gdal_config_options()].
#' @param platforms The names of Landsat satellites to download imagery from.
#' These do not correspond to the `platforms` column in [spectral_indices()];
#' the default argument of `c("landsat-9", "landsat-8")` corresponds to
#' the `Landsat-OLI` value in that column.
#'
#' @returns `output_filename`, unchanged.
#'
#' @examplesIf interactive()
#' aoi <- sf::st_point(c(-74.912131, 44.080410))
#' aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
#' aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)
#'
#' get_stac_data(aoi,
#'   start_date = "2022-06-01",
#'   end_date = "2022-06-30",
#'   pixel_x_size = 30,
#'   pixel_y_size = 30,
#'   asset_names = c(
#'     "red", "blue", "green"
#'   ),
#'   stac_source = "https://planetarycomputer.microsoft.com/api/stac/v1/",
#'   collection = "landsat-c2-l2",
#'   query_function = rsi_query_api,
#'   sign_function = sign_planetary_computer,
#'   mask_band = "qa_pixel",
#'   mask_function = landsat_mask_function,
#'   item_filter_function = landsat_platform_filter,
#'   platforms = c("landsat-9", "landsat-8"),
#'   output_filename = tempfile(fileext = ".tif")
#' )
#'
#' # or, mostly equivalently (will download more bands):
#' landsat_image <- get_landsat_imagery(
#'   aoi,
#'   start_date = "2022-06-01",
#'   end_date = "2022-08-30",
#'   output_filename = tempfile(fileext = ".tif")
#' )
#'
#' landsat_image <- terra::rast(landsat_image)
#' terra::plotRGB(terra::stretch(landsat_image))
#'
#' # The `get_*_imagery()` functions will download
#' # all available "data" assets by default
#' # (usually including measured values, and excluding derived bands)
#' sentinel1_data <- get_sentinel1_imagery(
#'   aoi,
#'   start_date = "2022-06-01",
#'   end_date = "2022-07-01",
#'   output_filename = tempfile(fileext = ".tif")
#' )
#' names(terra::rast(sentinel1_data))
#'
#' # You can see what bands will be downloaded by a function
#' # by inspecting the corresponding `band_mapping` object:
#' sentinel2_band_mapping$planetary_computer_v1
#'
#' # And you can add additional assets using `c()`:
#' c(
#'   sentinel2_band_mapping$planetary_computer_v1,
#'   "scl"
#' )
#'
#' # Or subset the assets downloaded using `[` or `[[`:
#' sentinel2_imagery <- get_sentinel2_imagery(
#'   aoi,
#'   start_date = "2022-06-01",
#'   end_date = "2022-07-01",
#'   asset_names = sentinel2_band_mapping$planetary_computer_v1["B01"],
#'   output_filename = tempfile(fileext = ".tif")
#' )
#' names(terra::rast(sentinel2_imagery))
#'
#' # If you're downloading data for a particularly large AOI,
#' # and can't composite the resulting images or want to make
#' # sure you can continue an interrupted download,
#' # consider tiling your AOI and requesting each tile separately:
#' aoi <- sf::st_make_grid(aoi, n = 2)
#' tiles <- lapply(
#'   seq_along(aoi),
#'   function(i) {
#'     get_landsat_imagery(
#'       aoi[i],
#'       start_date = "2022-06-01",
#'       end_date = "2022-08-30",
#'       output_filename = tempfile(fileext = ".tif")
#'     )
#'   }
#' )
#' # You'll get a list of tiles that you can then composite or
#' # work with however you wish:
#' unlist(tiles)
#'
#' @export
get_stac_data <- function(
  aoi,
  start_date,
  end_date,
  pixel_x_size = NULL,
  pixel_y_size = NULL,
  asset_names,
  stac_source,
  collection,
  ...,
  query_function = rsi_query_api,
  download_function = rsi_download_rasters,
  sign_function = NULL,
  rescale_bands = TRUE,
  item_filter_function = NULL,
  mask_band = NULL,
  mask_function = NULL,
  output_filename = paste0(proceduralnames::make_english_names(1), ".tif"),
  composite_function = c("merge", "median", "mean", "sum", "min", "max"),
  limit = 999,
  gdalwarp_options = rsi_gdalwarp_options(),
  gdal_config_options = rsi_gdal_config_options()
) {
  # query, filter, download, mask, composite, rescale
  if (!(inherits(aoi, "sf") || inherits(aoi, "sfc"))) {
    rlang::abort(
      "`aoi` must be an sf or sfc object.",
      class = "rsi_aoi_not_sf"
    )
  }

  aoi_bbox <- sf::st_bbox(aoi)
  if (aoi_bbox$xmin == aoi_bbox$xmax | aoi_bbox$ymin == aoi_bbox$ymax) {
    rlang::abort(
      c(
        "`aoi` has no extent.",
        "i" = "This can occur when `aoi` is a single point geometry."
      ),
      class = "rsi_aoi_is_point"
    )
  }

  if (
    sf::st_is_longlat(aoi) &&
      !(is.null(pixel_x_size) || is.null(pixel_y_size)) &&
      all(c(pixel_x_size, pixel_y_size) %in% c(10, 30))
  ) {
    rlang::warn(
      c(
        "The default pixel size arguments are intended for use with projected AOIs, but `aoi` appears to be in geographic coordinates.",
        i = glue::glue(
          "Pixel X size: {pixel_x_size}. Pixel Y size: {pixel_y_size}."
        ),
        i = glue::glue(
          "These dimensions will be interpreted in the same units as `aoi` (likely degrees), which may cause errors."
        )
      ),
      class = "rsi_default_pixel_size_geographic_coords"
    )
  }

  if (!is.null(composite_function)) {
    composite_function <- rlang::arg_match(composite_function)
  }

  if (is.null(mask_function)) {
    if (!is.null(mask_band)) {
      rlang::warn(
        c(
          "`mask_function` was NULL, but `mask_band` was not `NULL`.",
          i = "`mask_band` will be ignored (not downloaded or used)."
        ),
        class = "rsi_ignored_mask_band"
      )
    }
    mask_band <- NULL
  }

  check_type_and_length(
    start_date = character(1),
    end_date = character(1),
    pixel_x_size = numeric(1),
    pixel_y_size = numeric(1),
    stac_source = character(1),
    collection = character(1),
    rescale_bands = logical(1),
    mask_band = character(1),
    output_filename = character(1),
    composite_function = character(1),
    limit = numeric(1),
    gdalwarp_options = character()
  )

  if (is.null(sign_function) && is_pc(stac_source)) {
    sign_function <- sign_planetary_computer
  }

  gdalwarp_options <- process_gdalwarp_options(
    gdalwarp_options = gdalwarp_options,
    aoi = aoi,
    pixel_x_size = pixel_x_size,
    pixel_y_size = pixel_y_size
  )

  aoi_bbox <- sf::st_bbox(aoi)

  if (!is.null(start_date)) {
    start_date <- process_dates(start_date)
    end_date <- process_dates(end_date)
  }

  if (is.null(item_filter_function))
    item_filter_function <- function(x, ...) identity(x)

  # query
  items <- query_function(
    bbox = sf::st_bbox(sf::st_transform(aoi, 4326)),
    stac_source = stac_source,
    collection = collection,
    start_date = start_date,
    end_date = end_date,
    limit = limit,
    ...
  )
  # filter
  items <- item_filter_function(items, ...)

  if (!length(items$features)) {
    rlang::abort(
      "No items were found for this combination of collection, AOI, date range, and item filter function.",
      class = "rsi_no_items_found"
    )
  }

  if (missing(asset_names)) asset_names <- NULL
  if (is.null(asset_names)) {
    asset_names <- rstac::items_assets(items)
    if (length(asset_names) > 1) {
      rlang::warn(
        c(
          "`asset_names` was `NULL`, so rsi is attempting to download all assets in items in this collection.",
          i = "This includes multiple assets, so rsi is attempting to download all of them using the same download function.",
          i = "This might cause errors or not be what you want! Specify `asset_names` to fix this (and to silence this warning)."
        ),
        class = "rsi_missing_asset_names"
      )
    }
  }
  if (is.null(names(asset_names))) names(asset_names) <- asset_names

  items_urls <- extract_urls(asset_names, items)
  drop_mask_band <- FALSE
  if (!is.null(mask_band) && !(mask_band %in% names(items_urls))) {
    items_urls[[mask_band]] <- rstac::assets_url(items, mask_band)
    drop_mask_band <- TRUE
  }

  scale_strings <- character()
  if (rescale_bands) {
    scale_strings <- calc_scale_strings(names(items_urls), items)
  }
  if (length(scale_strings)) {
    scale_strings <- stats::setNames(
      paste("function(x) x", scale_strings),
      names(scale_strings)
    )
  } else {
    rescale_bands <- FALSE
  }

  merge_assets <- is.null(mask_function) &&
    !rescale_bands &&
    !is.null(composite_function) &&
    composite_function == "merge"

  # download
  # download_results is a data frame with names corresponding to "final" band
  # names and rows corresponding to individual STAC items
  download_results <- download_function(
    items = items,
    aoi = aoi_bbox,
    asset_names = stats::setNames(nm = names(items_urls)),
    sign_function = sign_function,
    merge = merge_assets,
    gdalwarp_options = gdalwarp_options,
    gdal_config_options = gdal_config_options,
    ...
  )
  if (!is.null(stats::na.action(download_results))) {
    items$features[stats::na.action(download_results)] <- NULL
  }

  temp_ras_files <- unlist(download_results)
  on.exit(file.remove(temp_ras_files), add = TRUE)
  # mask
  if (!is.null(mask_band)) {
    download_results <- rsi_apply_masks(
      download_locations = download_results,
      mask_band = mask_band,
      mask_function = mask_function
    )
  }

  download_results <- download_results[
    names(download_results) %in% names(asset_names)
  ]

  # composite
  output_vrt <- tempfile(fileext = ".vrt")
  if (is.null(composite_function)) {
    output_vrt <- replicate(nrow(download_results), tempfile(fileext = ".vrt"))
    # turn each row of the DF into its own character vector, stored in a list
    download_results <- apply(download_results, 1, identity, simplify = FALSE)
  } else if (!merge_assets) {
    download_results <- rsi_composite_bands(
      download_results,
      composite_function
    )
  } else {
    # turn DF into a character vector inside a list
    download_results <- list(unlist(download_results))
  }
  # rescale
  if (rescale_bands) {
    lapply(download_results, rescale_band, scale_strings)
  }

  if (drop_mask_band) items_urls[[mask_band]] <- NULL
  mapply(
    function(in_bands, vrt) {
      stack_rasters(
        in_bands[names(items_urls)],
        vrt,
        band_names = remap_band_names(names(items_urls), asset_names)
      )
    },
    in_bands = download_results,
    vrt = output_vrt
  )

  on.exit(file.remove(output_vrt), add = TRUE)

  if (is.null(composite_function)) {
    app <- tryCatch(rstac::items_datetime(items), error = function(e) NA)
    app <- gsub(":", "", app) # #29, #32
    if (any(is.na(app))) app <- NULL
    app <- app %||% as.character(seq_along(download_results))
    app <- make.unique(app, sep = "_")

    output_filename <- paste0(
      tools::file_path_sans_ext(output_filename),
      "_",
      app,
      ".",
      tools::file_ext(output_filename)
    )
  }

  out <- mapply(
    function(vrt, out) {
      sf::gdal_utils(
        "warp",
        vrt,
        out,
        options = gdalwarp_options
      )
      out
    },
    vrt = output_vrt,
    out = output_filename
  )

  # drop VRT filenames from vector
  as.vector(out)
}

#' @rdname get_stac_data
#' @export
get_sentinel1_imagery <- function(
  aoi,
  start_date,
  end_date,
  ...,
  pixel_x_size = 10,
  pixel_y_size = 10,
  asset_names = rsi::sentinel1_band_mapping$planetary_computer_v1,
  stac_source = attr(asset_names, "stac_source"),
  collection = attr(asset_names, "collection_name"),
  query_function = attr(asset_names, "query_function"),
  download_function = attr(asset_names, "download_function"),
  sign_function = attr(asset_names, "sign_function"),
  rescale_bands = FALSE,
  item_filter_function = NULL,
  mask_band = NULL,
  mask_function = NULL,
  output_filename = paste0(proceduralnames::make_english_names(1), ".tif"),
  composite_function = "median",
  limit = 999,
  gdalwarp_options = rsi_gdalwarp_options(),
  gdal_config_options = rsi_gdal_config_options()
) {
  args <- mget(names(formals()))
  args$`...` <- NULL
  args <- c(args, rlang::list2(...))
  do.call(get_stac_data, args)
}

#' @rdname get_stac_data
#' @export
get_sentinel2_imagery <- function(
  aoi,
  start_date,
  end_date,
  ...,
  pixel_x_size = 10,
  pixel_y_size = 10,
  asset_names = rsi::sentinel2_band_mapping$planetary_computer_v1,
  stac_source = attr(asset_names, "stac_source"),
  collection = attr(asset_names, "collection_name"),
  query_function = attr(asset_names, "query_function"),
  download_function = attr(asset_names, "download_function"),
  sign_function = attr(asset_names, "sign_function"),
  rescale_bands = FALSE,
  item_filter_function = NULL,
  mask_band = attr(asset_names, "mask_band"),
  mask_function = attr(asset_names, "mask_function"),
  output_filename = paste0(proceduralnames::make_english_names(1), ".tif"),
  composite_function = "median",
  limit = 999,
  gdalwarp_options = rsi_gdalwarp_options(),
  gdal_config_options = rsi_gdal_config_options()
) {
  args <- mget(names(formals()))
  args$`...` <- NULL
  args <- c(args, rlang::list2(...))
  do.call(get_stac_data, args)
}

#' @rdname get_stac_data
#' @export
get_landsat_imagery <- function(
  aoi,
  start_date,
  end_date,
  ...,
  platforms = c("landsat-9", "landsat-8"),
  pixel_x_size = 30,
  pixel_y_size = 30,
  asset_names = rsi::landsat_band_mapping$planetary_computer_v1,
  stac_source = attr(asset_names, "stac_source"),
  collection = attr(asset_names, "collection_name"),
  query_function = attr(asset_names, "query_function"),
  download_function = attr(asset_names, "download_function"),
  sign_function = attr(asset_names, "sign_function"),
  rescale_bands = TRUE,
  item_filter_function = landsat_platform_filter,
  mask_band = attr(asset_names, "mask_band"),
  mask_function = attr(asset_names, "mask_function"),
  output_filename = paste0(proceduralnames::make_english_names(1), ".tif"),
  composite_function = "median",
  limit = 999,
  gdalwarp_options = rsi_gdalwarp_options(),
  gdal_config_options = rsi_gdal_config_options()
) {
  args <- mget(names(formals()))
  args$`...` <- NULL
  args <- c(args, rlang::list2(...))
  do.call(get_stac_data, args)
}

#' @rdname get_stac_data
#' @export
get_naip_imagery <- function(
  aoi,
  start_date,
  end_date,
  ...,
  pixel_x_size = 1,
  pixel_y_size = 1,
  asset_names = "image",
  stac_source = "https://planetarycomputer.microsoft.com/api/stac/v1",
  collection = "naip",
  query_function = rsi_query_api,
  download_function = rsi_download_rasters,
  sign_function = sign_planetary_computer,
  rescale_bands = FALSE,
  output_filename = paste0(proceduralnames::make_english_names(1), ".tif"),
  composite_function = "merge",
  limit = 999,
  gdalwarp_options = rsi_gdalwarp_options(),
  gdal_config_options = rsi_gdal_config_options()
) {
  args <- mget(names(formals()))
  args$`...` <- NULL
  args <- c(args, rlang::list2(...))
  suppressWarnings(
    do.call(get_stac_data, args),
    classes = "rsi_band_name_length_mismatch"
  )
}

#' @rdname get_stac_data
#' @export
get_alos_palsar_imagery <- function(
  aoi,
  start_date,
  end_date,
  ...,
  pixel_x_size = 25,
  pixel_y_size = 25,
  asset_names = rsi::alos_palsar_band_mapping$planetary_computer_v1,
  stac_source = attr(asset_names, "stac_source"),
  collection = attr(asset_names, "collection_name"),
  query_function = attr(asset_names, "query_function"),
  download_function = attr(asset_names, "download_function"),
  sign_function = attr(asset_names, "sign_function"),
  rescale_bands = FALSE,
  item_filter_function = NULL,
  mask_band = attr(asset_names, "mask_band"),
  mask_function = attr(asset_names, "mask_function"),
  output_filename = paste0(proceduralnames::make_english_names(1), ".tif"),
  composite_function = "median",
  limit = 999,
  gdalwarp_options = rsi_gdalwarp_options(),
  gdal_config_options = rsi_gdal_config_options()
) {
  args <- mget(names(formals()))
  args$`...` <- NULL
  args <- c(args, rlang::list2(...))
  do.call(get_stac_data, args)
}

#' @rdname get_stac_data
#' @export
get_dem <- function(
  aoi,
  ...,
  start_date = NULL,
  end_date = NULL,
  pixel_x_size = 30,
  pixel_y_size = 30,
  asset_names = rsi::dem_band_mapping$planetary_computer_v1$`cop-dem-glo-30`,
  stac_source = attr(asset_names, "stac_source"),
  collection = attr(asset_names, "collection_name"),
  query_function = attr(asset_names, "query_function"),
  download_function = attr(asset_names, "download_function"),
  sign_function = attr(asset_names, "sign_function"),
  rescale_bands = FALSE,
  item_filter_function = NULL,
  mask_band = NULL,
  mask_function = NULL,
  output_filename = paste0(proceduralnames::make_english_names(1), ".tif"),
  composite_function = "max",
  limit = 999,
  gdalwarp_options = rsi_gdalwarp_options(),
  gdal_config_options = rsi_gdal_config_options()
) {
  args <- mget(names(formals()))
  args$`...` <- NULL
  args <- c(args, rlang::list2(...))
  do.call(get_stac_data, args)
}

rsi_apply_masks <- function(download_locations, mask_band, mask_function) {
  p <- build_progressr(
    nrow(download_locations) +
      (nrow(download_locations) * (ncol(download_locations) - 1))
  )

  apply(
    download_locations,
    1,
    function(files) {
      p("Running mask function")
      mask <- mask_function(terra::rast(files[[mask_band]]))

      lapply(
        files[setdiff(names(files), mask_band)],
        function(raster) {
          masked_file <- tempfile(fileext = ".tif")
          p("Applying mask to downloaded assets")
          terra::mask(
            terra::rast(raster),
            mask,
            maskvalues = c(NA, FALSE),
            filename = masked_file
          )
          file.rename(masked_file, raster)
          raster
        }
      )
    }
  )

  download_locations
}

rsi_composite_bands <- function(
  download_locations,
  composite_function = c("merge", "median", "mean", "sum", "min", "max")
) {
  composite_function <- rlang::arg_match(composite_function)

  p <- build_progressr(length(names(download_locations)))

  download_dir <- file.path(tempdir(), "composite_dir")
  if (!dir.exists(download_dir)) dir.create(download_dir)

  out <- vapply(
    names(download_locations),
    function(band_name) {
      p(glue::glue("Compositing {band_name}"))
      out_file <- file.path(download_dir, paste0(toupper(band_name), ".tif"))

      if (length(download_locations[[band_name]]) == 1) {
        file.copy(download_locations[[band_name]], out_file)
      } else if (composite_function == "merge") {
        do.call(
          terra::merge,
          list(
            x = terra::sprc(lapply(
              download_locations[[band_name]],
              terra::rast
            )),
            filename = out_file,
            overwrite = TRUE
          )
        )
      } else {
        utils::capture.output({
          do.call(
            terra::mosaic,
            list(
              x = terra::sprc(lapply(
                download_locations[[band_name]],
                terra::rast
              )),
              fun = composite_function,
              filename = out_file,
              overwrite = TRUE
            )
          )
        })
      }

      out_file
    },
    character(1)
  )

  list(out)
}
calc_scale_strings <- function(asset_names, items) {
  # Assign scale, offset attributes if they exist
  scales <- vapply(
    asset_names,
    get_rescaling_formula,
    items = items,
    element = "scale",
    numeric(1)
  )
  offsets <- vapply(
    asset_names,
    get_rescaling_formula,
    items = items,
    element = "offset",
    numeric(1)
  )

  scale_strings <- vapply(
    names(scales),
    function(band) {
      out <- ""

      if (band %in% names(scales) && !is.na(scales[[band]])) {
        out <- paste(out, glue::glue("* {scales[[band]]}"))
      }

      if (band %in% names(offsets) && !is.na(offsets[[band]])) {
        out <- paste(out, glue::glue("+ {offsets[[band]]}"))
      }

      out
    },
    character(1)
  )
  names(scale_strings) <- names(scales)

  scale_strings <- scale_strings[scale_strings != ""]
  scale_strings
}

rescale_band <- function(composited_bands, scale_strings) {
  p <- build_progressr(length(names(scale_strings)))

  for (band in names(scale_strings)) {
    p(glue::glue("Rescaling band {band}"))
    rescaled_file <- tempfile(fileext = ".tif")
    terra::writeRaster(
      eval(str2lang(scale_strings[[band]]))(terra::rast(composited_bands[[
        band
      ]])),
      filename = rescaled_file,
      overwrite = TRUE
    )
    file.rename(rescaled_file, composited_bands[[band]])
  }
  composited_bands
}

remap_band_names <- function(band_names, name_mapping) {
  ifelse(
    band_names %in% names(name_mapping),
    name_mapping[band_names],
    band_names
  )
}

process_gdalwarp_options <- function(
  gdalwarp_options,
  aoi,
  pixel_x_size = 30,
  pixel_y_size = 30
) {
  if (is.null(gdalwarp_options)) {
    gdalwarp_options <- character(0)
  }

  if (!("-t_srs" %in% gdalwarp_options)) {
    gdalwarp_options <- c(gdalwarp_options, "-t_srs", sf::st_crs(aoi)$wkt)
  }

  if (
    !("-tr" %in% gdalwarp_options) &&
      !is.null(pixel_x_size) &&
      !is.null(pixel_y_size)
  ) {
    gdalwarp_options <- c(gdalwarp_options, "-tr", pixel_x_size, pixel_y_size)
  }

  gdalwarp_options
}

process_dates <- function(date) {
  if (date == "..") {
    return(date)
  } # open intervals
  date <- as.POSIXct(date, "UTC")
  date <- strftime(date, "%Y-%m-%dT%H:%M:%S%Z", "UTC")
  gsub("UTC", "Z", date)
}

get_rescaling_formula <- function(items, band_name, element) {
  elements <- vapply(
    items$features,
    function(x) {
      x <- x$assets[[band_name]]$`raster:bands`[[1]]
      tryCatch(x[[element]] %||% NA_real_, error = function(e) NA_real_)
    },
    numeric(1)
  )

  if (length(unique(elements)) != 1) {
    rlang::warn(
      c(
        glue::glue("Images in band {band_name} have different {element}s."),
        i = "Returning images without rescaling."
      ),
      class = "rsi_multiple_scaling_formulas"
    )
    elements <- NA_real_
  }
  elements <- unique(elements)
  elements
}

is_pc <- function(url) {
  grepl("planetarycomputer.microsoft.com/api/stac/v1", url)
}

extract_urls <- function(asset_names, items) {
  items_urls <- lapply(
    names(asset_names),
    function(asset_name) suppressWarnings(rstac::assets_url(items, asset_name))
  )
  names(items_urls) <- names(asset_names)

  items_urls <- items_urls[!vapply(items_urls, is.null, logical(1))]

  items_urls
}

test_that("landsat_mask_function arguments work", {
  skip_on_cran()
  skip_if_offline()

  boston <- sf::st_point(c(-71.0610279, 42.361697)) |>
    sf::st_sfc(crs = 4326) |>
    sf::st_transform(2249) |>
    sf::st_buffer(10000)

  bands <- rsi::landsat_band_mapping$planetary_computer_v1[""]
  bands[[1]] <- "qa_pixel"
  names(bands) <- "qa_pixel"

  boston_landsat <- rsi::get_landsat_imagery(
    boston,
    "2022-06-01",
    "2022-07-01",
    asset_names = bands,
    output_filename = tempfile(fileext = ".tif"),
    mask_function = NULL,
    mask_band = NULL
  )

  expect_gt(
    sum(terra::values(landsat_mask_function(terra::rast(boston_landsat)))),
    sum(terra::values(landsat_mask_function(
      terra::rast(boston_landsat),
      "water"
    )))
  )

  expect_gt(
    sum(terra::values(landsat_mask_function(
      terra::rast(boston_landsat),
      "both"
    ))),
    sum(terra::values(landsat_mask_function(terra::rast(boston_landsat))))
  )

  expect_equal(
    sum(terra::values(landsat_mask_function(
      terra::rast(boston_landsat),
      "both"
    ))),
    sum(terra::values(landsat_mask_function(
      terra::rast(boston_landsat),
      c("land", "water")
    )))
  )

  expect_equal(
    sum(terra::values(landsat_mask_function(
      terra::rast(boston_landsat),
      "both"
    ))),
    sum(
      terra::values(
        landsat_mask_function(
          terra::rast(boston_landsat),
          masked_bits = list(
            clear_land = c(0:5, 7, 9, 11, 13, 15),
            clear_water = c(0:5, 9, 11, 13, 15)
          )
        )
      )
    )
  )
})

test_that("get_landsat_imagery() is stable", {
  skip_on_cran()
  skip_if_offline()
  aoi <- sf::st_point(c(-74.912131, 44.080410))
  aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
  aoi <- sf::st_buffer(sf::st_transform(aoi, 3857), 100)

  expect_no_error(
    out <- get_landsat_imagery(
      aoi,
      "2022-06-01",
      "2022-06-30",
      output_filename = tempfile(fileext = ".tif")
    )
  )
  expect_no_error(terra::rast(out))
  expect_true(
    all(
      names(terra::rast(out)) %in%
        as.vector(landsat_band_mapping$planetary_computer_v1)
    )
  )
})

test_that("get_sentinel1_imagery() is stable", {
  skip_on_cran()
  skip_if_offline()
  aoi <- sf::st_point(c(-74.912131, 44.080410))
  aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
  aoi <- sf::st_buffer(sf::st_transform(aoi, 3857), 100)

  expect_no_error(
    out <- get_sentinel1_imagery(
      aoi,
      "2022-06-01",
      "2022-06-30",
      output_filename = tempfile(fileext = ".tif")
    )
  )
  expect_no_error(terra::rast(out))
  expect_true(
    all(
      names(terra::rast(out)) %in%
        as.vector(sentinel1_band_mapping$planetary_computer_v1)
    )
  )
})

test_that("get_sentinel2_imagery() is stable", {
  skip_on_cran()
  skip_if_offline()
  aoi <- sf::st_point(c(-74.912131, 44.080410))
  aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
  aoi <- sf::st_buffer(sf::st_transform(aoi, 3857), 100)

  expect_no_error(
    out <- get_sentinel2_imagery(
      aoi,
      "2022-06-01",
      "2022-06-30",
      output_filename = tempfile(fileext = ".tif"),
      mask_function = sentinel2_mask_function
    )
  )
  expect_no_error(terra::rast(out))
  expect_true(
    all(
      names(terra::rast(out)) %in%
        as.vector(sentinel2_band_mapping$planetary_computer_v1)
    )
  )
})

test_that("get_dem() is stable", {
  skip_on_cran()
  skip_if_offline()
  aoi <- sf::st_point(c(-74.912131, 44.080410))
  aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
  aoi <- sf::st_buffer(sf::st_transform(aoi, 3857), 100)

  expect_no_error(
    out <- get_dem(
      aoi,
      output_filename = tempfile(fileext = ".tif")
    )
  )
  expect_no_error(terra::rast(out))
  expect_true(
    all(
      names(terra::rast(out)) %in%
        as.vector(dem_band_mapping$planetary_computer_v1)
    )
  )
})

test_that("non-default mappings work", {
  skip_on_cran()
  skip_if_offline()
  aoi <- sf::st_point(c(-74.912131, 44.080410))
  aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
  aoi <- sf::st_buffer(sf::st_transform(aoi, 3857), 100)

  expect_no_error(
    out <- get_sentinel2_imagery(
      aoi,
      "2022-06-01",
      "2022-06-30",
      output_filename = tempfile(fileext = ".tif"),
      asset_names = sentinel2_band_mapping$aws_v1
    )
  )
  expect_no_error(terra::rast(out))
  expect_true(
    all(
      names(terra::rast(out)) %in%
        as.vector(sentinel2_band_mapping$aws_v1)
    )
  )
})

test_that("can download RTC products", {
  skip_if(
    Sys.getenv("rsi_pc_key") == "",
    "Environment variable `rsi_pc_key` not set"
  )
  skip_on_cran()
  skip_if_offline()
  aoi <- sf::st_point(c(-74.912131, 44.080410))
  aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
  aoi <- sf::st_buffer(sf::st_transform(aoi, 3857), 100)

  expect_no_error(
    out <- get_sentinel1_imagery(
      aoi,
      "2022-06-01",
      "2022-06-30",
      output_filename = tempfile(fileext = ".tif"),
      collection = "sentinel-1-rtc"
    )
  )
  expect_no_error(terra::rast(out))
  expect_true(
    all(
      names(terra::rast(out)) %in%
        as.vector(sentinel1_band_mapping$planetary_computer_v1)
    )
  )
})

test_that("hidden arguments work", {
  skip_on_cran()
  skip_if_offline()
  aoi <- sf::st_point(c(-74.912131, 44.080410))
  aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
  aoi <- sf::st_buffer(sf::st_transform(aoi, 3857), 100)

  expect_no_error(
    out <- get_landsat_imagery(
      aoi,
      "2022-06-01",
      "2022-06-30",
      output_filename = tempfile(fileext = ".tif"),
      query_function = rsi_query_api,
      mask_function = landsat_mask_function
    )
  )
  expect_no_error(terra::rast(out))
  expect_true(
    all(
      names(terra::rast(out)) %in%
        as.vector(landsat_band_mapping$planetary_computer_v1)
    )
  )
})

test_that("simple merge method works", {
  skip_on_cran()
  skip_if_offline()
  aoi <- sf::st_point(c(-74.912131, 44.080410))
  aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
  aoi <- sf::st_buffer(sf::st_transform(aoi, 3857), 100)

  expect_no_error(
    out <- get_stac_data(
      aoi,
      "2021-01-01",
      "2021-12-31",
      asset_names = "lcpri",
      stac_source = "https://planetarycomputer.microsoft.com/api/stac/v1",
      collection = "usgs-lcmap-conus-v13",
      query_function = rsi_query_api,
      sign_function = rsi::sign_planetary_computer,
      output_filename = tempfile(fileext = ".tif")
    )
  )
  expect_no_error(terra::rast(out))
})

test_that("warning (but not error) fires if `mask_band` is not NULL with NULL `mask_function`", {
  skip_on_cran()
  skip_if_offline()
  aoi <- sf::st_point(c(-74.912131, 44.080410))
  aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
  aoi <- sf::st_buffer(sf::st_transform(aoi, 3857), 100)

  expect_warning(
    x <- get_landsat_imagery(
      aoi = aoi,
      start_date = "2022-06-01",
      end_date = "2022-08-01",
      mask_function = NULL,
      rescale_bands = FALSE,
      output_filename = tempfile(fileext = ".tif")
    ),
    class = "rsi_ignored_mask_band"
  )
})

test_that("get_*_data works with mapply() (#17)", {
  skip_on_cran()
  skip_if_offline()
  san_antonio <- sf::st_point(c(-98.491142, 29.424349))
  san_antonio <- sf::st_sfc(san_antonio, crs = "EPSG:4326")
  san_antonio <- sf::st_buffer(sf::st_transform(san_antonio, "EPSG:3081"), 100)

  expect_no_error(
    mapply(
      get_landsat_imagery,
      start_date = c("2023-09-01", "2023-10-01"),
      end_date = c("2023-09-30", "2023-10-31"),
      output_filename = replicate(2, tempfile(fileext = ".tif")),
      MoreArgs = c(aoi = list(san_antonio))
    )
  )
})

test_that("proper error if no items are found", {
  skip_if_offline()
  skip_on_cran()

  aoi <- sf::st_point(c(-74.912131, 44.080410))
  aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
  aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)

  expect_error(
    get_stac_data(
      aoi,
      start_date = "1970-01-01", # pre-LCMAP dates
      end_date = "1970-12-31",
      asset_names = "lcpri",
      stac_source = "https://planetarycomputer.microsoft.com/api/stac/v1/",
      collection = "usgs-lcmap-conus-v13",
      output_filename = tempfile(fileext = ".tif"),
    ),
    class = "rsi_no_items_found"
  )
})

test_that("no-composite paths work on Windows #29, #32", {
  skip_if_offline()
  skip_on_cran()
  skip_on_os("mac")
  skip_on_os("linux")

  aoi <- sf::st_buffer(
    sf::st_transform(
      sf::st_sfc(sf::st_point(c(-74.912131, 44.080410)), crs = 4326),
      3857
    ),
    100
  )

  expect_no_error(
    get_landsat_imagery(
      aoi = aoi,
      start_date = "2022-06-01",
      end_date = "2022-08-01",
      composite_function = NULL,
      output_filename = tempfile(fileext = ".tif")
    )
  )
})

test_that("no-composites return the same data", {
  skip_if_offline()
  skip_on_cran()

  aoi <- sf::st_buffer(
    sf::st_transform(
      sf::st_sfc(sf::st_point(c(-74.912131, 44.080410)), crs = 4326),
      3857
    ),
    100
  )

  expect_contains(
    names(
      terra::rast(
        get_landsat_imagery(
          aoi = aoi,
          start_date = "2022-07-01",
          end_date = "2022-07-05",
          composite_function = NULL,
          output_filename = tempfile(fileext = ".tif")
        )
      )
    ),
    setdiff(rsi::landsat_band_mapping$planetary_computer_v1, "T")
  )
})

test_that("get_naip_imagery() is stable", {
  skip_on_cran()
  skip_if_offline()
  aoi <- sf::st_point(c(-74.912131, 44.080410))
  aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
  aoi <- sf::st_buffer(sf::st_transform(aoi, 3857), 100)

  expect_no_error(
    out <- get_naip_imagery(
      aoi,
      "2018-01-01",
      "2020-01-31",
      output_filename = tempfile(fileext = ".tif")
    )
  )
  expect_no_error(terra::rast(out))
})

test_that("get_alos_palsar_imagery() is stable", {
  skip_on_cran()
  skip_if_offline()
  aoi <- sf::st_point(c(-74.912131, 44.080410))
  aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
  aoi <- sf::st_buffer(sf::st_transform(aoi, 3857), 100)

  expect_no_error(
    out <- get_alos_palsar_imagery(
      aoi,
      "2021-01-01",
      "2021-12-31",
      output_filename = tempfile(fileext = ".tif")
    )
  )
  expect_no_error(terra::rast(out))
})

test_that("non-sf AOI throws an error", {
  expect_error(
    get_stac_data(2),
    class = "rsi_aoi_not_sf"
  )
})

test_that("Providing pixel sizes with geographic coords fires the expected warning", {
  skip_on_cran()
  skip_if_offline()
  aoi <- sf::st_point(c(-74.912131, 44.080410))
  aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)

  expect_error(
    tryCatch(
      get_stac_data(aoi, pixel_x_size = 30, pixel_y_size = 30),
      rsi_default_pixel_size_geographic_coords = rlang::abort(
        "The warning fired",
        class = "success"
      )
    ),
    class = "success"
  )
})

test_that("Providing no asset names fires the expected warning", {
  skip_on_cran()
  skip_if_offline()
  aoi <- sf::st_point(c(-74.912131, 44.080410))
  aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
  aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)

  expect_error(
    tryCatch(
      get_stac_data(
        aoi,
        start_date = "2022-01-01",
        end_date = "2022-12-31",
        stac_source = "https://planetarycomputer.microsoft.com/api/stac/v1/",
        collection = "usgs-lcmap-conus-v13",
        output_filename = tempfile(fileext = ".tif"),
      ),
      rsi_missing_asset_names = rlang::abort(
        "The warning fired",
        class = "success"
      ),
      class = "success"
    )
  )
})

test_that("Providing a single point geometry throws an error", {
  expect_error(
    get_landsat_imagery(
      aoi = sf::st_sfc(sf::st_point(c(-74.912131, 44.080410)), crs = 4326),
      start_date = "2022-06-01",
      end_date = "2022-08-01",
      composite_function = NULL,
      output_filename = tempfile(fileext = ".tif")
    ),
    class = "rsi_aoi_is_point"
  )
})

#' Create and save a multi-band output raster by combining input rasters
#'
#' This function creates an output raster that "stacks" all the bands of its
#' input rasters, as though they were loaded one after another into a GIS. It
#' does this by first constructing a GDAL virtual raster, or "VRT", and then
#' optionally uses GDAL's warper to convert this VRT into a standalone file.
#' The VRT is fast to create and does not require much space, but does require
#' the input rasters not be moved or altered. Creating a standalone raster from
#' this file may take a long time and a large amount of disk space.
#'
#' @param rasters A list of rasters to combine into a single multi-band raster,
#' as character file paths to files that can be read by [terra::rast()].
#' Rasters will be "stacked" upon one another, preserving values. They must
#' share CRS.
#' @param output_filename The location to save the final "stacked" raster. If
#' this filename has a "vrt" extension as determined by `tools::file_ext()`,
#' then this function exits after creating a VRT; otherwise, this function will
#' create a VRT and then use `sf::gdal_utils("warp")` to convert the VRT into
#' another format.
#' @inheritParams rlang::args_dots_empty
#' @param resolution Numeric of length 2, representing the target X and Y
#' resolution of the output raster. If only a single value is provided, it will
#' be used for both X and Y resolution; if more than 2 values are provided, an
#' error is thrown.
#' @param extent Numeric of length 4, representing the target xmin, ymin, xmax,
#' and ymax values of the output raster (its bounding box), in that order.
#' @param reference_raster The position (index) of the raster in `rasters` to
#' take extent, resolution, and CRS information from. No reprojection is done.
#' If `resolution` or `extent` are provided, they override the values from the
#' reference raster.
#' @param resampling_method The method to use when resampling to different
#' resolutions in the VRT.
#' @param band_names Either a character vector of band names, or a function that
#' when given a character vector of band names, returns a character vector of
#' the same length containing new band names.
#' @param check_crs Logical: Should this function check that all `rasters` share
#' the same CRS? Set to `FALSE` only if you are entirely confident that rasters
#' have equivalent CRS definitions, but not identical WKT strings.
#' @param gdalwarp_options Options passed to `gdalwarp` through the `options`
#' argument of [sf::gdal_utils()]. This argument is ignored (with a warning)
#' if `output_filename` is a VRT.
#' @param gdal_config_options Options passed to `gdalwarp` through the
#' `config_options` argument of [sf::gdal_utils()].  This argument is ignored
#' (with a warning) if `output_filename` is a VRT.
#'
#' @returns `output_filename`, unchanged.
#'
#' @examples
#' stack_rasters(
#'   list(
#'     system.file("rasters/dpdd.tif", package = "rsi"),
#'     system.file("rasters/example_sentinel1.tif", package = "rsi")
#'   ),
#'   tempfile(fileext = ".vrt")
#' )
#'
#' @export
stack_rasters <- function(
  rasters,
  output_filename,
  ...,
  resolution,
  extent,
  reference_raster = 1,
  resampling_method = "bilinear",
  band_names,
  check_crs = TRUE,
  gdalwarp_options = c(
    "-multi",
    "-overwrite",
    "-co",
    "COMPRESS=DEFLATE",
    "-co",
    "PREDICTOR=2",
    "-co",
    "NUM_THREADS=ALL_CPUS"
  ),
  gdal_config_options = c(
    VSI_CACHE = "TRUE",
    GDAL_CACHEMAX = "30%",
    VSI_CACHE_SIZE = "10000000",
    GDAL_HTTP_MULTIPLEX = "YES",
    GDAL_INGESTED_BYTES_AT_OPEN = "32000",
    GDAL_DISABLE_READDIR_ON_OPEN = "EMPTY_DIR",
    GDAL_HTTP_VERSION = "2",
    GDAL_HTTP_MERGE_CONSECUTIVE_RANGES = "YES",
    GDAL_NUM_THREADS = "ALL_CPUS"
  )
) {
  rlang::check_dots_empty()

  check_type_and_length(
    output_filename = character(1),
    resampling_method = character(1)
  )

  tryCatch(
    check_type_and_length(rasters = list()),
    error = function(e) {
      check_type_and_length(
        rasters = character(),
        call = rlang::caller_env(4)
      )
    }
  )

  out_dir <- dirname(output_filename)

  use_warper <- tolower(tools::file_ext(output_filename)) != "vrt"

  if (
    !use_warper && (!missing(gdal_config_options) || !missing(gdalwarp_options))
  ) {
    rlang::warn(
      "`gdal_config_options` and `gdalwarp_options` are both ignored when `output_filename` ends in 'vrt'.",
      class = "rsi_gdal_options_ignored"
    )
  }

  if (
    !(reference_raster %in%
      seq_along(rasters) ||
      reference_raster %in% names(rasters))
  ) {
    if (is.numeric(reference_raster)) {
      msg <- glue::glue(
        "`rasters` is of length {length(rasters)}, but `reference_raster` is {reference_raster}."
      )
    } else {
      msg <- glue::glue(
        "`reference_raster` is '{reference_raster}', but none of the elements in `rasters` are named '{reference_raster}'."
      )
    }
    rlang::abort(
      c(
        "`reference_raster` must be a valid index for `rasters`",
        i = msg
      ),
      class = "rsi_not_in_vec"
    )
  }

  ref_rast <- terra::rast(rasters[[reference_raster]])

  if (missing(extent)) {
    extent <- terra::ext(ref_rast)
    extent <- as.vector(extent)[c("xmin", "ymin", "xmax", "ymax")]
    names(extent) <- NULL
  } else {
    if (length(extent) != 4) {
      rlang::abort(
        c(
          "`extent` must be 4 numbers (in xmin ymin xmax ymax order).",
          i = glue::glue("{length(extent)} values were provided.")
        ),
        class = "rsi_bad_extent"
      )
    }
  }

  if (missing(resolution)) {
    resolution <- terra::res(ref_rast)
  } else {
    if (length(resolution) == 1) resolution <- c(resolution, resolution)
    if (length(resolution) != 2) {
      rlang::abort(
        c(
          "`resolution` must be 2 numbers (in xres yres order).",
          i = glue::glue("{length(resolution)} values were provided.")
        ),
        class = "rsi_bad_resolution"
      )
    }
  }

  ref_crs <- terra::crs(ref_rast)

  var_names <- unlist(
    lapply(
      rasters,
      function(r) {
        r <- terra::rast(r)
        # this is the only place we instantiate these rasters, so may as well
        # check CRS alignment while we're here...
        if (check_crs && (terra::crs(r) != ref_crs)) {
          rlang::abort(
            c(
              "Rasters do not all share the reference raster's CRS.",
              i = "Reproject rasters to all share the same CRS."
            ),
            class = "rsi_multiple_crs",
            call = rlang::caller_env()
          )
        }
        names(r)
      }
    )
  )

  if (!missing(band_names) && is.function(band_names)) {
    var_names <- band_names(var_names)
  }
  # technically we could have skipped reading the names of each raster here
  # and just accepted the band names, but we want to check the CRS anyway
  if (!missing(band_names) && is.character(band_names)) {
    var_names <- band_names
  }

  vrt_container_file <- tempfile(tmpdir = out_dir, fileext = ".vrt")
  sf::gdal_utils(
    "buildvrt",
    rasters[[reference_raster]],
    vrt_container_file,
    options = c(
      "-b",
      1,
      "-te",
      extent,
      "-tr",
      resolution,
      "-r",
      resampling_method
    )
  )
  vrt_container <- readLines(vrt_container_file)
  file.remove(vrt_container_file)

  intermediate_vrt <- lapply(
    rasters,
    function(raster) {
      r_lyrs <- terra::nlyr(terra::rast(raster))
      out_files <- replicate(
        r_lyrs,
        tempfile(tmpdir = out_dir, fileext = ".vrt")
      )

      lapply(
        seq_len(r_lyrs),
        function(b) {
          sf::gdal_utils(
            "buildvrt",
            raster,
            out_files[[b]],
            options = c(
              "-b",
              b,
              "-te",
              extent,
              "-tr",
              resolution,
              "-r",
              resampling_method
            )
          )
        }
      )
      out_files
    }
  )

  intermediate_vrt <- unlist(intermediate_vrt)
  on.exit(file.remove(intermediate_vrt), add = TRUE)

  band_no <- 1
  vrt_bands <- vector("list", length(intermediate_vrt))

  set_names <- FALSE
  if (length(intermediate_vrt) == length(var_names)) {
    set_names <- TRUE
  } else if (!missing(band_names)) {
    rlang::warn(
      c(
        "The number of band names provided did not match the number of bands in the output file.",
        i = "Ignoring the provided band names and using default values instead."
      ),
      class = "rsi_band_name_length_mismatch"
    )
  }

  for (vrt in intermediate_vrt) {
    vrt <- readLines(vrt)
    band_def <- grep("VRTRasterBand", vrt)
    vrt <- vrt[seq(band_def[[1]], band_def[[2]])]
    vrt[1] <- gsub("band=\"1\"", paste0("band=\"", band_no, "\""), vrt[1])
    if (set_names) {
      vrt <- c(
        vrt[1],
        paste0("    <Description>", var_names[[band_no]], "</Description>"),
        vrt[2:length(vrt)]
      )
    }

    vrt_bands[[band_no]] <- vrt

    band_no <- band_no + 1
  }

  vrt_destination <- ifelse(
    use_warper,
    tempfile(fileext = ".vrt"),
    output_filename
  )

  band_def <- grep("VRTRasterBand", vrt_container)
  writeLines(
    c(
      vrt_container[1:(band_def[[1]] - 1)],
      unlist(vrt_bands),
      vrt_container[(band_def[[2]] + 1):length(vrt_container)]
    ),
    vrt_destination
  )

  if (use_warper) {
    sf::gdal_utils(
      "warp",
      source = vrt_destination,
      destination = output_filename,
      options = gdalwarp_options,
      config_options = gdal_config_options
    )
  }

  output_filename
}

test_that("stack_rasters works", {
  expect_equal(
    terra::values(
      terra::rast(
        stack_rasters(
          list(
            system.file("rasters/example_sentinel1.tif", package = "rsi")
          ),
          tempfile(fileext = ".vrt")
        )
      )
    ),
    terra::values(terra::rast(system.file(
      "rasters/example_sentinel1.tif",
      package = "rsi"
    )))
  )
})

test_that("stack_rasters works with non-VRT outputs", {
  expect_no_error(
    out_tif <- stack_rasters(
      list(
        system.file("rasters/example_sentinel1.tif", package = "rsi")
      ),
      tempfile(fileext = ".tif")
    )
  )

  # the re-compression means we don't expect this to necessarily be the same size
  # but it should still be a decent sized file, not just a text file
  expect_true(
    file.info(out_tif)$size >
      (file.info(system.file(
        "rasters/example_sentinel1.tif",
        package = "rsi"
      ))$size /
        2)
  )

  expect_equal(
    terra::values(terra::rast(out_tif, drivers = "GTiff")),
    terra::values(terra::rast(system.file(
      "rasters/example_sentinel1.tif",
      package = "rsi"
    )))
  )
})

test_that("stack_rasters fails when reference_raster isn't in the vector", {
  expect_error(
    stack_rasters(
      list(
        system.file("rasters/example_sentinel1.tif", package = "rsi")
      ),
      tempfile(fileext = ".vrt"),
      reference_raster = 2
    ),
    class = "rsi_not_in_vec"
  )

  expect_error(
    stack_rasters(
      list(
        system.file("rasters/example_sentinel1.tif", package = "rsi")
      ),
      tempfile(fileext = ".vrt"),
      reference_raster = "some_raster"
    ),
    class = "rsi_not_in_vec"
  )
})

test_that("stack_rasters fails when extent isn't four numbers", {
  expect_error(
    stack_rasters(
      list(
        system.file("rasters/example_sentinel1.tif", package = "rsi")
      ),
      tempfile(fileext = ".vrt"),
      extent = 20
    ),
    class = "rsi_bad_extent"
  )
})

test_that("stack_rasters fails when resolution isn't 1-2 numbers", {
  expect_error(
    stack_rasters(
      list(
        system.file("rasters/example_sentinel1.tif", package = "rsi")
      ),
      tempfile(fileext = ".vrt"),
      resolution = c(20, 30, 40)
    ),
    class = "rsi_bad_resolution"
  )
})

test_that("stack_rasters fails when rasters don't share a CRS", {
  s1 <- tempfile(fileext = ".tif")
  terra::writeRaster(
    terra::project(
      terra::rast(
        system.file("rasters/example_sentinel1.tif", package = "rsi")
      ),
      "EPSG:4326"
    ),
    s1
  )

  expect_error(
    stack_rasters(
      list(
        system.file("rasters/example_sentinel1.tif", package = "rsi"),
        s1
      ),
      tempfile(fileext = ".vrt")
    ),
    class = "rsi_multiple_crs"
  )
})

test_that("stack_rasters fails when rasters are not character vectors", {
  r1 <- terra::rast(matrix(rnorm(100), 10))
  expect_snapshot(
    stack_rasters(r1, "a"),
    error = TRUE
  )
})

test_that("stack_rasters warns when arguments are being ignored", {
  expect_warning(
    stack_rasters(
      list(
        system.file("rasters/example_sentinel1.tif", package = "rsi")
      ),
      tempfile(fileext = ".vrt"),
      gdalwarp_options = c("THIS IS IGNORED")
    ),
    class = "rsi_gdal_options_ignored"
  )
})

test_that("type_and_length checks", {
  expect_snapshot(
    stack_rasters("a", c("a", "b")),
    error = TRUE
  )

  expect_snapshot(
    stack_rasters("a", "b", resampling_method = c("a", "b")),
    error = TRUE
  )
})

test_that("stack_rasters warns when band names aren't the right length", {
  expect_warning(
    stack_rasters(
      list(system.file("rasters/example_sentinel1.tif", package = "rsi")),
      tempfile(fileext = ".vrt"),
      band_names = "VH"
    ),
    class = "rsi_band_name_length_mismatch"
  )

  expect_warning(
    stack_rasters(
      list(system.file("rasters/example_sentinel1.tif", package = "rsi")),
      tempfile(fileext = ".vrt"),
      band_names = c("VH", "VV", "EXTRA")
    ),
    class = "rsi_band_name_length_mismatch"
  )
})

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
## usethis namespace: end
NULL

#' @importFrom rlang %||%
#' @importFrom stats predict
#' @importFrom httr user_agent

utils::globalVariables(c(
  "dem_band_mapping",
  "landsat_band_mapping",
  "sentinel1_band_mapping",
  "sentinel2_band_mapping",
  "short_names"
))

#' Create a Sentinel-2 mask raster from the SCL band
#'
#' @param raster The SCL band of a Sentinel-2 image
#'
#' @returns A boolean raster to be used to mask a Sentinel-2 image
#'
#' @examplesIf interactive()
#' aoi <- sf::st_point(c(-74.912131, 44.080410))
#' aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
#' aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)
#'
#' sentinel2_image <- get_sentinel2_imagery(
#'   aoi,
#'   start_date = "2022-06-01",
#'   end_date = "2022-08-30",
#'   mask_function = sentinel2_mask_function
#' )
#'
#' @export
sentinel2_mask_function <- function(raster) {
  terra::`%in%`(
    raster,
    c(
      4, # VEGETATION
      5, # NOT_VEGETATED
      6, # WATER
      7, # UNCLASSIFIED
      11 # SNOW
    )
  )
  # That means we drop:
  # c(
  # 0, # NO_DATA
  # 1, # SATURATED_OR_DEFECTIVE
  # 2, # DARK_AREA_PIXELS
  # 3, # CLOUD_SHADOWS
  # 8, # CLOUD_MEDIUM_PROBABILITY
  # 9, # CLOUD_HIGH_PROBABILITY
  # 10 # THIN_CIRRUS
  # )
}

#' Create a Landsat mask raster from the QA band
#'
#' @param raster The QA band of a Landsat image
#' @param include Include pixels that represent land, water, or both? Passing
#' `c("land", "water")` is identical to passing `"both"`.
#' @inheritParams rlang::args_dots_empty
#' @param masked_bits Optionally, a list of integer vectors representing the
#' individual bits to mask out. Each vector is converted to an integer
#' representation, and then pixels with matching `qa_pixel` values
#' are preserved by the mask. Refer to the Landsat science product guide for
#' further information on what bit values represent for your platform of
#' interest.
#'
#' @returns A boolean raster to be used to mask a Landsat image
#'
#' @examplesIf interactive()
#' aoi <- sf::st_point(c(-74.912131, 44.080410))
#' aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
#' aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)
#'
#' landsat_image <- get_landsat_imagery(
#'   aoi,
#'   start_date = "2022-06-01",
#'   end_date = "2022-08-30",
#'   mask_function = landsat_mask_function,
#'   output_file = tempfile(fileext = ".tif")
#' )
#'
#' # Or, optionally pass the qa_pixel bits to mask out directly
#' landsat_image <- get_landsat_imagery(
#'   aoi,
#'   start_date = "2022-06-01",
#'   end_date = "2022-08-30",
#'   mask_function = \(x) landsat_mask_function(
#'     x,
#'     masked_bits = list(c(0:5, 7, 9, 11, 13, 15))
#'   ),
#'   output_file = tempfile(fileext = ".tif")
#' )
#'
#' # You can use this to specify multiple acceptable values
#' # from the qa_pixel bitmask; names are optional
#' landsat_image <- get_landsat_imagery(
#'   aoi,
#'   start_date = "2022-06-01",
#'   end_date = "2022-08-30",
#'   mask_function = \(x) landsat_mask_function(
#'     x,
#'     masked_bits = list(
#'       clear_land = c(0:5, 7, 9, 11, 13, 15),
#'       clear_water = c(0:5, 9, 11, 13, 15)
#'     )
#'   ),
#'   output_file = tempfile(fileext = ".tif")
#' )
#'
#' @export
landsat_mask_function <- function(
  raster,
  include = c("land", "water", "both"),
  ...,
  masked_bits
) {
  rlang::check_dots_empty()
  if (missing(masked_bits)) {
    if (missing(include)) include <- include[[1]]
    include <- rlang::arg_match(include, multiple = TRUE)

    masked_bits <- list()
    if (any(c("land", "both") %in% include)) {
      masked_bits <- c(masked_bits, list(c(0:5, 7, 9, 11, 13, 15)))
    }
    if (any(c("water", "both") %in% include)) {
      masked_bits <- c(masked_bits, list(c(0:5, 9, 11, 13, 15)))
    }
  } else if (!missing(include)) {
    rlang::abort(
      "Only one of `include` and `masked_bits` can be specified.",
      class = "rsi_masked_bits_and_include"
    )
  }

  classes <- vapply(masked_bits, bits_to_int, integer(1), USE.NAMES = FALSE)

  terra::`%in%`(raster, classes)
}

bits_to_int <- function(vals) {
  bits <- integer(16)
  bits[setdiff(0:15, vals) + 1] <- 1
  strtoi(paste0(rev(bits), collapse = ""), 2)
}

#' Create an ALOS PALSAR mask raster from the mask band
#'
#' @param raster The mask band of an ALOS PALSAR image
#' @param include Include pixels that represent land, water, or both? Passing
#' `c("land", "water")` is identical to passing `"both"`.
#'
#' @returns A boolean raster to be used to mask an ALOS PALSAR image
#'
#' @examplesIf interactive()
#' aoi <- sf::st_point(c(-74.912131, 44.080410))
#' aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
#' aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)
#'
#' palsar_image <- get_alos_palsar_imagery(
#'   aoi,
#'   start_date = "2021-01-01",
#'   end_date = "2021-12-31",
#'   mask_function = alos_palsar_mask_function,
#'   output_file = tempfile(fileext = ".tif"),
#'   gdalwarp_options = c(
#'     rsi::rsi_gdalwarp_options(),
#'     "-srcnodata", "nan"
#'   )
#' )
#'
#' @export
alos_palsar_mask_function <- function(
  raster,
  include = c("land", "water", "both")
) {
  if (missing(include)) include <- include[[1]]
  include <- rlang::arg_match(include, multiple = TRUE)
  # includes no data (0), layover (100) and shadow (150)
  # https://docs.digitalearthafrica.org/en/latest/sandbox/notebooks/Datasets/ALOS_PALSAR_Annual_Mosaic.html
  classes <- c(0, 100, 150)

  if (any(c("land", "both") %in% include)) classes <- c(classes, 255)
  if (any(c("water", "both") %in% include)) classes <- c(classes, 50)

  terra::`%in%`(raster, classes)
}

#' Deprecated functions
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' These functions have been deprecated in favor of better approaches.
#'
#' * `default_query_function()` was renamed to `rsi_query_api()`. These
#'   functions are identical, and the older name will be removed in a future
#'   release.
#'
#' @name deprecated
#' @keywords internal
#' @export
default_query_function <- function(
  bbox,
  stac_source,
  collection,
  start_date,
  end_date,
  limit,
  ...
) {
  lifecycle::deprecate_warn(
    "0.2.0",
    "default_query_function()",
    "rsi_query_api()"
  )
  rsi_query_api(
    bbox = bbox,
    stac_source = stac_source,
    collection = collection,
    start_date = start_date,
    end_date = end_date,
    limit = limit,
    ...
  )
}

check_type_and_length <- function(
  ...,
  call = rlang::caller_env(),
  env = rlang::caller_env()
) {
  dots <- list(...)
  if (length(dots) == 0) {
    return(invisible(TRUE))
  }

  if (is.null(names(dots)) || any(names(dots) == "")) {
    rlang::abort(
      "All arguments to `check_type_and_length()` must be named.",
      class = "rsi_unnamed_check_args",
      call = call
    )
  }

  problem_args <- character()
  for (dot in names(dots)) {
    arg <- get(dot, envir = env)
    if (is.null(arg)) next

    arg_class <- class(arg)
    dot_class <- class(dots[[dot]])
    if (!any(arg_class %in% dot_class)) {
      if ("integer" %in% arg_class && "numeric" %in% dot_class) {
        next # Purposefully doing nothing -- rely on implicit conversion
      } else if ("integer" %in% dot_class && rlang::is_integerish(arg)) {
        next # Purposefully doing nothing -- rely on implicit conversion
      } else {
        problem_args <- c(
          problem_args,
          glue::glue(
            "{dot} should be a {class(dots[[dot]])}, but is a {class(arg)}."
          )
        )
      }
    }
    if (length(dots[[dot]]) && length(arg) != length(dots[[dot]])) {
      problem_args <- c(
        problem_args,
        glue::glue(
          "{dot} should be of length {length(dots[[dot]])}, but is length {length(arg)}."
        )
      )
    }
  }

  if (length(problem_args)) {
    names(problem_args) <- rep("*", length(problem_args))
    rlang::abort(
      c(
        "Some input arguments weren't the right class or length:",
        problem_args
      ),
      call = call,
      class = "rsi_incorrect_type_or_length"
    )
  }
  return(invisible(TRUE))
}

test_that("check_type_and_length works", {
  expect_true(check_type_and_length())
  expect_error(
    check_type_and_length(2),
    class = "rsi_unnamed_check_args"
  )
  f1 <- function(x = 2L) {
    check_type_and_length(x = character(1))
  }
  expect_error(
    f1(),
    class = "rsi_incorrect_type_or_length"
  )
})

test_that("integers are valid numerics", {
  f1 <- function(x = 2L) {
    check_type_and_length(x = numeric(1))
  }
  expect_true(f1())
})

test_that("integer-ish are valid integers", {
  f1 <- function(x = 2) {
    check_type_and_length(x = integer(1))
  }
  expect_true(f1())
})

test_that("multi-class arguments are fine", {
  f1 <- function(x) {
    check_type_and_length(x = character())
  }
  y <- "a"
  class(y) <- c("glue", class(y))

  expect_true(f1(y))
})

#' Sentinel-2 band mapping
#'
#' This object is a named list of character vectors, with names corresponding to
#' Sentinel-2 band names and values corresponding to band names in
#' `spectral_indices`.
#'
#' @template band_mappings
"sentinel2_band_mapping"

#' Sentinel-1 band mapping
#'
#' This object is a named list of character vectors, with names corresponding to
#' Sentinel-1 band names and values corresponding to band names in
#' `spectral_indices`.
#'
#' @template band_mappings
"sentinel1_band_mapping"

#' Landsat band mapping
#'
#' This object is a named list of character vectors, with names corresponding to
#' Landsat band names and values corresponding to band names in
#' `spectral_indices`.
#'
#' @template band_mappings
"landsat_band_mapping"

#' ALOS PALSAR band mapping
#'
#' This object is a named list of character vectors, with names corresponding to
#' Landsat band names and values corresponding to band names in
#' `spectral_indices`.
#'
#' @template band_mappings
"alos_palsar_band_mapping"


#' Landsat band mapping
#'
#' This object is structured slightly differently from other band mapping
#' objects; it is a list of named lists, whose names correspond to DEM
#' collections available within a given STAC catalog. Those named lists are
#' then more standard band mapping objects, containing character vectors with
#' names corresponding to asset names and values equal to `elevation`.
#'
#' @template band_mappings
"dem_band_mapping"

#' Calculate indices from the bands of a raster
#'
#' This function computes any number of indices from an input raster via
#' [terra::predict()]. By default, this function is designed to work with
#' subsets of [spectral_indices()], but it will work with any data frame with a
#' `formula`, `bands`, and `short_name` column.
#'
#' @section Security:
#' Note that this function is running code from the `formula` column of the
#' spectral indices data frame, which is derived from a JSON file downloaded off
#' the internet. It's not impossible that an attacker could take advantage of
#' this to run arbitrary code on your computer. To mitigate this, indices are
#' calculated in a minimal environment that contains very few functions or
#' symbols (preventing an attacker from accessing, for example, `system()`).
#'
#' Still, it's good practice to inspect your `formula` column to make sure
#' there's nothing nasty hiding in any of the formulas you're going to run.
#' Additionally, consider using pre-saved indices tables or
#' `spectral_indices(download_indices = FALSE)` if using this in an unsupervised
#' workload.
#'
#' @param raster The raster (either as a SpatRaster or object readable by
#' [terra::rast()]) to compute indices from.
#' @param indices A data frame of indices to compute. The intent is for this
#' function to work with subsets of [spectral_indices()], but you can provide
#' any data frame with `formula`, `bands`, and `short_name` columns; see
#' "Custom indices" below.
#' @param output_filename The filename to write the computed metrics to.
#' @inheritParams rlang::args_dots_empty
#' @inheritParams terra::predict
#' @param extra_objects A named list of additional objects to pass to the
#' minimal environment that formulas are executed in. For instance, if you
#' need to use the `pmax` function in order to calculate an index, you can
#' make it available in the environment by setting
#' `extra_objects = list("pmax" = pmax)`. Providing extra functionality is
#' inherently less safe than the default minimal environment, and as such always
#' emits a warning, which you can suppress with `suppressWarnings()`.
#' @param names_suffix If not `NULL`, will be used (with [paste()]) to add a
#' suffix to each of the band names returned.
#'
#' @section Custom indices:
#' While this function is designed to work with subsets of [spectral_indices()],
#' you can also provide your own custom indices by providing a data frame with
#' at least three columns:
#'
#' * `formula` should contain a string representation of the equation used to
#' calculate the index. This string will be interpreted as R code, and should use
#' normal R arithmetic operators. Any additional functions used in calculating the
#' index will need to be passed through `extra_objects`.
#' * `bands` should be a list column containing character vectors identifying the
#' bands required to calculate the index. These vectors will be checked against the
#' band names of `raster` to make sure that all necessary bands are available
#' before indices are calculated.
#' * `short_name` should contain a string which will be used as the band name in
#' the final output file.
#'
#' Sometimes you might want to tweak the formula used for an existing index, for
#' instance to set non-data parameters to constant values. In that case, make sure
#' you also update `bands` to reflect just the bands required by your updated formula,
#' so that the function won't error unnecessarily.
#'
#' @return `output_filename`, unchanged.
#'
#' @examples
#' our_raster <- system.file("rasters/example_sentinel1.tif", package = "rsi")
#' calculate_indices(
#'   our_raster,
#'   filter_bands(bands = names(terra::rast(our_raster))),
#'   tempfile(fileext = ".tif"),
#'   names_suffix = "sentinel1"
#' )
#'
#' # Formulas aren't able to access most R functions or operators,
#' # in order to try and keep formulas from doing something bad:
#' example_indices <- filter_platforms(platforms = "Sentinel-1 (Dual Polarisation VV-VH)")[1, ]
#' example_indices$formula <- 'system("echo something bad")'
#' # So this will error:
#' try(
#'   calculate_indices(
#'     system.file("rasters/example_sentinel1.tif", package = "rsi"),
#'     example_indices,
#'     tempfile(fileext = ".tif")
#'   )
#' )
#'
#' # Because of this, formulas which try to use most R functions
#' # will wind up erroring as well:
#' example_indices$formula <- "pmax(VH, VV)"
#' try(
#'   calculate_indices(
#'     system.file("rasters/example_sentinel1.tif", package = "rsi"),
#'     example_indices,
#'     tempfile(fileext = ".tif")
#'   )
#' )
#'
#' # To fix this, pass the objects you want to use to `extra_objects`
#' calculate_indices(
#'   system.file("rasters/example_sentinel1.tif", package = "rsi"),
#'   example_indices,
#'   tempfile(fileext = ".tif"),
#'   extra_objects = list(pmax = pmax)
#' ) |>
#'   suppressWarnings(classes = "rsi_extra_objects")
#'
#' @export
calculate_indices <- function(
  raster,
  indices,
  output_filename,
  ...,
  cores = 1L,
  wopt = list(),
  overwrite = FALSE,
  extra_objects = list(),
  names_suffix = NULL
) {
  rlang::check_dots_empty()
  rlang::check_installed("terra")
  if (!all(c("formula", "short_name") %in% names(indices))) {
    rlang::abort(
      "Both `formula` and `short_name` must be columns in `indices`.",
      class = "rsi_missing_column"
    )
  }

  if (!inherits(raster, "SpatRaster")) raster <- terra::rast(raster)
  check_indices(names(raster), indices)

  formulas <- lapply(indices[["formula"]], str2lang)
  paste_names <- !is.null(names_suffix) && names_suffix != ""

  exec_objects <- list(
    # math functions
    `-` = `-`,
    `(` = `(`,
    `*` = `*`,
    `/` = `/`,
    `^` = `^`,
    `+` = `+`,
    # necessary syntax
    `<-` = `<-`,
    `if` = `if`,
    `{` = `{`,
    `function` = `function`,
    # renaming
    names = names,
    `names<-` = `names<-`,
    paste = paste,
    # pieces for predicting
    predict = terra::predict,
    list = list(),
    lapply = lapply,
    with = with,
    eval = eval,
    # user-provided variables
    formulas = formulas,
    short_names = indices[["short_name"]],
    paste_names = paste_names,
    raster = raster,
    output_filename = output_filename,
    wopt = wopt,
    cores = cores,
    overwrite = overwrite,
    names_suffix = names_suffix
  )

  if (length(extra_objects)) {
    rlang::warn(
      c(
        "Providing extra objects can potentially make it easier for malicious code to impact your system.",
        i = "Make sure you closely inspect the formulas you're running, before running them, and understand why they need extra objects!",
        i = "This warning can be silenced using `suppressWarnings(..., classes = 'rsi_extra_objects')`"
      ),
      class = "rsi_extra_objects"
    )
    exec_objects <- c(
      exec_objects,
      extra_objects
    )
  }

  exec_env <- rlang::new_environment(exec_objects)

  # covr can't instrument inside of our locked-down environment
  # so either we widen the environment (which seems like the wrong thing to do)
  # or we ignore this chunk wrt coverage
  # nocov start
  local(
    {
      predict(
        raster,
        list,
        fun = function(model, newdata) {
          out <- lapply(
            formulas,
            function(calc) {
              with(newdata, eval(calc))
            }
          )
          names(out) <- short_names
          if (paste_names) {
            names(out) <- paste(names(out), names_suffix, sep = "_")
          }
          out
        },
        filename = output_filename,
        wopt = wopt,
        cores = cores,
        overwrite = overwrite
      )
    },
    envir = exec_env
  )
  # nocov end

  output_filename
}

check_indices <- function(
  remap_band_names,
  indices,
  call = rlang::caller_env()
) {
  good <- TRUE
  if (!is.null(remap_band_names)) {
    good <- vapply(
      indices$bands,
      function(bands) all(bands %in% unlist(remap_band_names)),
      logical(1)
    )
  }

  if (!any(good)) {
    rlang::abort(
      "Some indices cannot be calculated using the available image bands.",
      call = call,
      class = "rsi_missing_indices"
    )
  }
}

#' Filter indices based on (relatively) complicated fields
#'
#' @param platforms,bands Names of the instruments (for `platforms`) or spectra
#' (for `bands`) indices must contain.
#' @param indices The data frame to filter. Must contain the relevant column.
#' @param operand A function defining how to apply this filter.
#' For instance, `operand = all` means that the index must contain all the
#' `platforms` or `bands` provided, while `operand = any` means that the index
#' must contain at least one of the `platforms` or `bands` provided.
#' @param type What type of query is this? If `filter`, then indices are
#' returned if all/any the bands they use (depending on `operand`) are in
#' `bands`. If `search`, then indices are returned if all/any of `bands` are in
#' the bands they use.
#'
#' @examples
#' filter_platforms(platforms = "Sentinel-2")
#' filter_platforms(platforms = c("Landsat-OLI", "Sentinel-2"))
#' filter_bands(bands = c("R", "N"), operand = any)
#'
#' @returns A filtered version of `indices`.
#'
#' @rdname filters
#' @export
filter_platforms <- function(
  indices = spectral_indices(),
  platforms = unique(unlist(
    spectral_indices(download_indices = FALSE, update_cache = FALSE)$platforms
  )),
  operand = c("all", "any")
) {
  platforms <- rlang::arg_match(platforms, multiple = TRUE)
  if (missing(operand)) {
    operand <- rlang::arg_match(operand)
  }
  ret_indices <- vapply(
    indices$platforms,
    function(x) rlang::exec(operand, platforms %in% x),
    logical(1)
  )
  indices[ret_indices, , drop = FALSE]
}

#' @rdname filters
#' @export
filter_bands <- function(
  indices = spectral_indices(),
  bands = unique(unlist(
    spectral_indices(download_indices = FALSE, update_cache = FALSE)$bands
  )),
  operand = c("all", "any"),
  type = c("filter", "search")
) {
  bands <- rlang::arg_match(bands, multiple = TRUE)
  type <- rlang::arg_match(type)
  if (missing(operand)) {
    operand <- rlang::arg_match(operand)
  }
  ret_indices <- vapply(
    indices$bands,
    function(x) {
      if (type == "search") {
        rlang::exec(operand, bands %in% x)
      } else {
        rlang::exec(operand, x %in% bands)
      }
    },
    logical(1)
  )
  indices[ret_indices, , drop = FALSE]
}

test_that("filter_platforms() works", {
  # Get package indices, in case we're on CRAN without internet:
  idxs <- suppressWarnings(
    spectral_indices(update_cache = FALSE),
    classes = "rsi_failed_download"
  )

  test_plats <- c("Landsat-OLI", "Sentinel-2")

  expect_no_error(
    # no internet and no cache file == warning
    both_plats <- suppressWarnings(
      filter_platforms(idxs, test_plats),
      classes = "rsi_failed_download"
    )
  )

  expect_true(
    all(
      vapply(
        both_plats$platforms,
        function(x) all(test_plats %in% x),
        logical(1)
      )
    )
  )

  expect_no_error(
    either_plat <- suppressWarnings(
      filter_platforms(idxs, test_plats, operand = "any"),
      classes = "rsi_failed_download"
    )
  )

  expect_true(
    all(
      vapply(
        either_plat$platforms,
        function(x) any(test_plats %in% x),
        logical(1)
      )
    )
  )
})

test_that("filter_bands() works", {
  # Get package indices, in case we're on CRAN without internet:
  idxs <- suppressWarnings(
    spectral_indices(update_cache = FALSE),
    classes = "rsi_failed_download"
  )

  test_bands <- c("R", "N")

  expect_no_error(
    both_bands <- suppressWarnings(
      filter_bands(idxs, test_bands),
      classes = "rsi_failed_download"
    )
  )

  expect_true(
    all(
      vapply(both_bands$bands, function(x) all(x %in% test_bands), logical(1))
    )
  )

  expect_no_error(
    either_band <- suppressWarnings(
      filter_bands(idxs, test_bands, operand = "any"),
      classes = "rsi_failed_download"
    )
  )

  expect_true(
    all(
      vapply(either_band$bands, function(x) any(test_bands %in% x), logical(1))
    )
  )
})

#' Filter Landsat features to only specific platforms
#'
#' @param items A `STACItemCatalog` containing some number of features
#' @param platforms A vector of acceptable platforms, for instance `landsat-9`.
#' Note that this refers to satellite names, and _not_ to platforms in
#' `spectral_indices()`.
#'
#' @returns A `STACItemCollection`.
#'
#' @examplesIf interactive()
#' aoi <- sf::st_point(c(-74.912131, 44.080410))
#' aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
#' aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)
#'
#' landsat_image <- get_landsat_imagery(
#'   aoi,
#'   start_date = "2022-06-01",
#'   end_date = "2022-08-30",
#'   item_filter_function = landsat_platform_filter
#' )
#'
#' @export
landsat_platform_filter <- function(items, platforms) {
  acceptable_platforms <- vapply(
    items$features,
    function(x)
      tryCatch(x$properties$platform %in% platforms, error = function(e) FALSE),
    logical(1)
  )
  items$features <- items$features[acceptable_platforms]
  items
}

test_that("subsetting works", {
  expect_snapshot(landsat_band_mapping$planetary_computer_v1)
  expect_snapshot(landsat_band_mapping$planetary_computer_v1["red"])
  expect_snapshot(landsat_band_mapping$planetary_computer_v1[["red"]])
  expect_snapshot(
    landsat_band_mapping$planetary_computer_v1[
      landsat_band_mapping$planetary_computer_v1 == "R"
    ]
  )
})

test_that("c works", {
  expect_snapshot(c(
    rsi::sentinel2_band_mapping$planetary_computer_v1,
    scl = "scl"
  ))
  expect_identical(
    c(rsi::sentinel2_band_mapping$planetary_computer_v1, scl = "scl"),
    c(rsi::sentinel2_band_mapping$planetary_computer_v1, "scl")
  )
})

---
title: "A whirlwind tour of rsi"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{rsi}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = rlang::is_installed("curl") && !is.null(curl::nslookup("r-project.org", error = FALSE)) && !(!interactive() && !isTRUE(as.logical(Sys.getenv("NOT_CRAN", "false"))))
)
```

The rsi package aims to help you download data from STAC APIs, to compute spectral indices from data, and to handle a handful of other spatial data wrangling tasks. This vignette walks through the main features of the package. Let's start off by loading the package:

```{r setup}
library(rsi)
```

## Downloading data

In order to download data from a STAC API, we're going to need to specify both a spatial and temporal area of interest. To define our spatial area of interest, we'll first mock up a small bounding box in northern Massachusetts (in a planar CRS appropriate for the region):

```{r}
our_aoi <- sf::st_bbox(
  c(xmin = 200000, ymin = 900000, xmax = 200100, ymax = 900100),
  crs = 26986
)
our_aoi <- sf::st_as_sf(sf::st_as_sfc(our_aoi))
sf::st_area(our_aoi)
plot(sf::st_geometry(our_aoi))
```

This is the area we're going to download and process data for.

Out of the box, rsi includes a number of functions to make it easy to download the most popular data sets from public STAC APIs. For instance, the `get_landsat_imagery()` has a number of default arguments for downloading Landsat 8 and 9 imagery from Microsoft's Planetary Computer. We'll pass our spatial area of interest along with a timeframe to this function, in order to get a cloud-masked composite image of all Landsat acquisitions for our spatiotemporal window. Let's grab all the imagery from September 2023:

```{r}
our_imagery <- get_landsat_imagery(
  our_aoi,
  "2023-09-01",
  "2023-09-30",
  output_filename = tempfile(fileext = ".tif")
)
our_imagery
```

If we wanted to, we could use `future::plan()` here to specify a parallelization methodology to speed up downloads by using multiple threads. We could also use `progressr::handlers()` to get a progress bar to report our download status.

By default, `get_landsat_imagery()` will download a composite of all bands available in Landsat 8 and 9 imagery for our timeframe:

```{r}
terra::rast(our_imagery) |>
  terra::plot()
```

Notice that this composite has been cloud-masked (using the QA pixel band) and rescaled (using the scale and offset specified in metadata provided by the STAC endpoint) automatically. You can control these behaviors via function arguments.

We're also able to use rsi to download other data sets -- for instance, we could also grab a DEM for this area from Planetary Computer, using the `get_dem()` function:

```{r}
our_dem <- get_dem(our_aoi)
terra::rast(our_dem) |>
  terra::plot()
```

Under the hood, both of these functions (and their friends, `get_sentinel2_imagery()` and `get_sentinel1_imagery()`) are powered by a lower-level `get_stac_data()` function, which should theoretically work with any imagery provided by any STAC API, anywhere. These functions simply provide user-friendly defaults to make it faster to get the data you care about.

## Finding spectral indices

In addition to these STAC-focused data-downloading functions, rsi also has an interface to the [Awesome Spectral Indices](https://github.com/awesome-spectral-indices/awesome-spectral-indices) project via the `spectral_indices()` function:

```{r}
spectral_indices() |>
  head()
```

This function attempts to grab the newest version of the spectral indices JSON file from the ASI repo, and then stores that data in a cache folder on your computer. If the downloading fails, the package will fall back (with a warning) to use your possibly outdated cache instead; if you don't have a cache and can't download the files, the package will instead (with a different warning) resort to using a packaged version of the indices file. This ensures that you're always getting the latest and greatest version of the ASI list possible, but that the package can still be used without an internet connection.

There are also functions in rsi to sort through the ASI list of indices. For instance, the `filter_platforms()` function can be used to, well, filter the list to only indices that can be calculated from a given platform. For instance, to filter to only indices that can be calculated using data from Landsat's Operational Land Imager:

```{r}
filter_platforms(platforms = "Landsat-OLI") |>
  head()
```

There's an equivalent function to filter indices based upon the bands that require. For instance, we can filter the list to only indices that use the red and blue band of images:

```{r}
filter_bands(bands = c("R", "B"))
```

Arguments to these functions let you control whether you're looking for indices that match _all_ platforms and bands you've specified, or _any_ of them.

## Calculating indices

But rsi doesn't simply make these formulas available in R, it also helps you compute these indices from imagery, via the `calculate_indices()` function. This function takes your imagery and a subset of `spectral_indices()` as arguments, and creates a raster containing all of those indices as an output. We can use `filter_bands()` to quickly get a list of the indices we can compute from our Landsat imagery, and then `calculate_indices()` to compute all 128 of those indices from our images:

```{r}
our_indices <- calculate_indices(
  our_imagery,
  filter_bands(bands = names(terra::rast(our_imagery))),
  "our_indices.tif"
)
terra::rast(our_indices) |>
  terra::plot()
```

Note that `calculate_indices()` is evaluating the formulas in the `formula` column of the spectral indices data frame as if they were code. These formulas are evaluated inside of a very limited environment, which doesn't have access to the global environment or most R fixtures, which does _reduce_ the amount of harm malicious code could do:

```{r}
evil_index <- spectral_indices()[1, ]
evil_index$formula <- "base::system('echo OHNO')"
try(
  calculate_indices(
    our_imagery,
    evil_index,
    tempfile(fileext = ".tif")
  )
)
```

But it's worth scanning your formulas before running `calculate_indices()`, just to make sure you aren't going to be accidentally running something surprising!

## Stacking rasters

Last but not least, rsi also provides a way to combine disparate data sets covering the same geographic region into a single VRT, quickly creating a file that you can treat as a single raster without taking up much additional storage space. This is a great way to create predictor bricks from your indices and downloaded data which you can then use for model fitting and prediction:

```{r}
combined_layers <- stack_rasters(
  c(our_imagery, our_dem, our_indices),
  tempfile(fileext = ".vrt")
)

terra::rast(combined_layers) |>
  terra::plot()
```

# subsetting works

    Code
      landsat_band_mapping$planetary_computer_v1
    Output
      An rsi band mapping object with attributes:
      names mask_band mask_function stac_source collection_name query_function download_function sign_function class
      
      coastal    blue   green     red   nir08  swir16  swir22    lwir  lwir11 
          "A"     "B"     "G"     "R"     "N"    "S1"    "S2"     "T"    "T1" 

---

    Code
      landsat_band_mapping$planetary_computer_v1["red"]
    Output
      An rsi band mapping object with attributes:
      mask_band mask_function stac_source collection_name query_function download_function sign_function class names
      
      red 
      "R" 

---

    Code
      landsat_band_mapping$planetary_computer_v1[["red"]]
    Output
      An rsi band mapping object with attributes:
      mask_band mask_function stac_source collection_name query_function download_function sign_function class names
      
      red 
      "R" 

---

    Code
      landsat_band_mapping$planetary_computer_v1[landsat_band_mapping$
        planetary_computer_v1 == "R"]
    Output
      An rsi band mapping object with attributes:
      mask_band mask_function stac_source collection_name query_function download_function sign_function class names
      
      red 
      "R" 

# c works

    Code
      c(rsi::sentinel2_band_mapping$planetary_computer_v1, scl = "scl")
    Output
      An rsi band mapping object with attributes:
      mask_band mask_function stac_source collection_name query_function class scl_name download_function sign_function names
      
        B01   B02   B03   B04   B05   B06   B07   B08   B8A   B09   B11   B12   scl 
        "A"   "B"   "G"   "R" "RE1" "RE2" "RE3"   "N"  "N2"  "WV"  "S1"  "S2" "scl" 

# stack_rasters fails when rasters are not character vectors

    Code
      stack_rasters(r1, "a")
    Condition
      Error in `stack_rasters()`:
      ! Some input arguments weren't the right class or length:
      * rasters should be a character, but is a SpatRaster.

# type_and_length checks

    Code
      stack_rasters("a", c("a", "b"))
    Condition
      Error in `stack_rasters()`:
      ! Some input arguments weren't the right class or length:
      * output_filename should be of length 1, but is length 2.

---

    Code
      stack_rasters("a", "b", resampling_method = c("a", "b"))
    Condition
      Error in `stack_rasters()`:
      ! Some input arguments weren't the right class or length:
      * resampling_method should be of length 1, but is length 2.

# no cache, update false

    Code
      suppressWarnings(spectral_indices(update_cache = FALSE))
    Output
      # A tibble: 271 x 9
         application_domain bands     contributor   date_of_addition formula long_name
         <chr>              <list>    <chr>         <chr>            <chr>   <chr>    
       1 vegetation         <chr [2]> https://gith~ 2021-11-17       (N - 0~ Aerosol ~
       2 vegetation         <chr [2]> https://gith~ 2021-11-17       (N - 0~ Aerosol ~
       3 water              <chr [6]> https://gith~ 2022-09-22       (B + G~ Augmente~
       4 vegetation         <chr [2]> https://gith~ 2021-09-20       (1 / G~ Anthocya~
       5 vegetation         <chr [3]> https://gith~ 2022-04-08       N * ((~ Anthocya~
       6 vegetation         <chr [4]> https://gith~ 2021-05-11       (N - (~ Atmosphe~
       7 vegetation         <chr [4]> https://gith~ 2021-05-14       sla * ~ Adjusted~
       8 vegetation         <chr [2]> https://gith~ 2022-04-08       (N * (~ Advanced~
       9 water              <chr [4]> https://gith~ 2021-09-18       4.0 * ~ Automate~
      10 water              <chr [5]> https://gith~ 2021-09-18       B + 2.~ Automate~
      # i 261 more rows
      # i 3 more variables: platforms <list>, reference <chr>, short_name <chr>

# no cache, download false

    Code
      spectral_indices(download_indices = FALSE)
    Condition
      Warning:
      No cache file present and `download_indices` set to `FALSE`.
      i Returning (likely outdated) package data instead.
    Output
      # A tibble: 271 x 9
         application_domain bands     contributor   date_of_addition formula long_name
         <chr>              <list>    <chr>         <chr>            <chr>   <chr>    
       1 vegetation         <chr [2]> https://gith~ 2021-11-17       (N - 0~ Aerosol ~
       2 vegetation         <chr [2]> https://gith~ 2021-11-17       (N - 0~ Aerosol ~
       3 water              <chr [6]> https://gith~ 2022-09-22       (B + G~ Augmente~
       4 vegetation         <chr [2]> https://gith~ 2021-09-20       (1 / G~ Anthocya~
       5 vegetation         <chr [3]> https://gith~ 2022-04-08       N * ((~ Anthocya~
       6 vegetation         <chr [4]> https://gith~ 2021-05-11       (N - (~ Atmosphe~
       7 vegetation         <chr [4]> https://gith~ 2021-05-14       sla * ~ Adjusted~
       8 vegetation         <chr [2]> https://gith~ 2022-04-08       (N * (~ Advanced~
       9 water              <chr [4]> https://gith~ 2021-09-18       4.0 * ~ Automate~
      10 water              <chr [5]> https://gith~ 2021-09-18       B + 2.~ Automate~
      # i 261 more rows
      # i 3 more variables: platforms <list>, reference <chr>, short_name <chr>

# no cache, download and update false

    Code
      spectral_indices(download_indices = FALSE, update_cache = FALSE)
    Condition
      Warning:
      No cache file present and `download_indices` set to `FALSE`.
      i Returning (likely outdated) package data instead.
    Output
      # A tibble: 271 x 9
         application_domain bands     contributor   date_of_addition formula long_name
         <chr>              <list>    <chr>         <chr>            <chr>   <chr>    
       1 vegetation         <chr [2]> https://gith~ 2021-11-17       (N - 0~ Aerosol ~
       2 vegetation         <chr [2]> https://gith~ 2021-11-17       (N - 0~ Aerosol ~
       3 water              <chr [6]> https://gith~ 2022-09-22       (B + G~ Augmente~
       4 vegetation         <chr [2]> https://gith~ 2021-09-20       (1 / G~ Anthocya~
       5 vegetation         <chr [3]> https://gith~ 2022-04-08       N * ((~ Anthocya~
       6 vegetation         <chr [4]> https://gith~ 2021-05-11       (N - (~ Atmosphe~
       7 vegetation         <chr [4]> https://gith~ 2021-05-14       sla * ~ Adjusted~
       8 vegetation         <chr [2]> https://gith~ 2022-04-08       (N * (~ Advanced~
       9 water              <chr [4]> https://gith~ 2021-09-18       4.0 * ~ Automate~
      10 water              <chr [5]> https://gith~ 2021-09-18       B + 2.~ Automate~
      # i 261 more rows
      # i 3 more variables: platforms <list>, reference <chr>, short_name <chr>

# download false and update true

    Code
      spectral_indices(download_indices = FALSE, update_cache = TRUE)
    Condition
      Error in `spectral_indices()`:
      ! Cannot update the cache if not downloading indices.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{rsi_gdal_options}
\alias{rsi_gdal_options}
\alias{rsi_gdal_config_options}
\alias{rsi_gdalwarp_options}
\title{Default options for GDAL}
\usage{
rsi_gdal_config_options()

rsi_gdalwarp_options()
}
\value{
A vector of options for GDAL commands.
}
\description{
These functions provide useful default options for GDAL functions,
making downloading and warping (hopefully!) more efficient for
most use cases.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mask_functions.R
\name{sentinel2_mask_function}
\alias{sentinel2_mask_function}
\title{Create a Sentinel-2 mask raster from the SCL band}
\usage{
sentinel2_mask_function(raster)
}
\arguments{
\item{raster}{The SCL band of a Sentinel-2 image}
}
\value{
A boolean raster to be used to mask a Sentinel-2 image
}
\description{
Create a Sentinel-2 mask raster from the SCL band
}
\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
aoi <- sf::st_point(c(-74.912131, 44.080410))
aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)

sentinel2_image <- get_sentinel2_imagery(
  aoi,
  start_date = "2022-06-01",
  end_date = "2022-08-30",
  mask_function = sentinel2_mask_function
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/landsat_platform_filter.R
\name{landsat_platform_filter}
\alias{landsat_platform_filter}
\title{Filter Landsat features to only specific platforms}
\usage{
landsat_platform_filter(items, platforms)
}
\arguments{
\item{items}{A \code{STACItemCatalog} containing some number of features}

\item{platforms}{A vector of acceptable platforms, for instance \code{landsat-9}.
Note that this refers to satellite names, and \emph{not} to platforms in
\code{spectral_indices()}.}
}
\value{
A \code{STACItemCollection}.
}
\description{
Filter Landsat features to only specific platforms
}
\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
aoi <- sf::st_point(c(-74.912131, 44.080410))
aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)

landsat_image <- get_landsat_imagery(
  aoi,
  start_date = "2022-06-01",
  end_date = "2022-08-30",
  item_filter_function = landsat_platform_filter
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_stac_data.R
\name{get_stac_data}
\alias{get_stac_data}
\alias{get_sentinel1_imagery}
\alias{get_sentinel2_imagery}
\alias{get_landsat_imagery}
\alias{get_naip_imagery}
\alias{get_alos_palsar_imagery}
\alias{get_dem}
\title{Retrieve raster data from STAC endpoints}
\usage{
get_stac_data(
  aoi,
  start_date,
  end_date,
  pixel_x_size = NULL,
  pixel_y_size = NULL,
  asset_names,
  stac_source,
  collection,
  ...,
  query_function = rsi_query_api,
  download_function = rsi_download_rasters,
  sign_function = NULL,
  rescale_bands = TRUE,
  item_filter_function = NULL,
  mask_band = NULL,
  mask_function = NULL,
  output_filename = paste0(proceduralnames::make_english_names(1), ".tif"),
  composite_function = c("merge", "median", "mean", "sum", "min", "max"),
  limit = 999,
  gdalwarp_options = rsi_gdalwarp_options(),
  gdal_config_options = rsi_gdal_config_options()
)

get_sentinel1_imagery(
  aoi,
  start_date,
  end_date,
  ...,
  pixel_x_size = 10,
  pixel_y_size = 10,
  asset_names = rsi::sentinel1_band_mapping$planetary_computer_v1,
  stac_source = attr(asset_names, "stac_source"),
  collection = attr(asset_names, "collection_name"),
  query_function = attr(asset_names, "query_function"),
  download_function = attr(asset_names, "download_function"),
  sign_function = attr(asset_names, "sign_function"),
  rescale_bands = FALSE,
  item_filter_function = NULL,
  mask_band = NULL,
  mask_function = NULL,
  output_filename = paste0(proceduralnames::make_english_names(1), ".tif"),
  composite_function = "median",
  limit = 999,
  gdalwarp_options = rsi_gdalwarp_options(),
  gdal_config_options = rsi_gdal_config_options()
)

get_sentinel2_imagery(
  aoi,
  start_date,
  end_date,
  ...,
  pixel_x_size = 10,
  pixel_y_size = 10,
  asset_names = rsi::sentinel2_band_mapping$planetary_computer_v1,
  stac_source = attr(asset_names, "stac_source"),
  collection = attr(asset_names, "collection_name"),
  query_function = attr(asset_names, "query_function"),
  download_function = attr(asset_names, "download_function"),
  sign_function = attr(asset_names, "sign_function"),
  rescale_bands = FALSE,
  item_filter_function = NULL,
  mask_band = attr(asset_names, "mask_band"),
  mask_function = attr(asset_names, "mask_function"),
  output_filename = paste0(proceduralnames::make_english_names(1), ".tif"),
  composite_function = "median",
  limit = 999,
  gdalwarp_options = rsi_gdalwarp_options(),
  gdal_config_options = rsi_gdal_config_options()
)

get_landsat_imagery(
  aoi,
  start_date,
  end_date,
  ...,
  platforms = c("landsat-9", "landsat-8"),
  pixel_x_size = 30,
  pixel_y_size = 30,
  asset_names = rsi::landsat_band_mapping$planetary_computer_v1,
  stac_source = attr(asset_names, "stac_source"),
  collection = attr(asset_names, "collection_name"),
  query_function = attr(asset_names, "query_function"),
  download_function = attr(asset_names, "download_function"),
  sign_function = attr(asset_names, "sign_function"),
  rescale_bands = TRUE,
  item_filter_function = landsat_platform_filter,
  mask_band = attr(asset_names, "mask_band"),
  mask_function = attr(asset_names, "mask_function"),
  output_filename = paste0(proceduralnames::make_english_names(1), ".tif"),
  composite_function = "median",
  limit = 999,
  gdalwarp_options = rsi_gdalwarp_options(),
  gdal_config_options = rsi_gdal_config_options()
)

get_naip_imagery(
  aoi,
  start_date,
  end_date,
  ...,
  pixel_x_size = 1,
  pixel_y_size = 1,
  asset_names = "image",
  stac_source = "https://planetarycomputer.microsoft.com/api/stac/v1",
  collection = "naip",
  query_function = rsi_query_api,
  download_function = rsi_download_rasters,
  sign_function = sign_planetary_computer,
  rescale_bands = FALSE,
  output_filename = paste0(proceduralnames::make_english_names(1), ".tif"),
  composite_function = "merge",
  limit = 999,
  gdalwarp_options = rsi_gdalwarp_options(),
  gdal_config_options = rsi_gdal_config_options()
)

get_alos_palsar_imagery(
  aoi,
  start_date,
  end_date,
  ...,
  pixel_x_size = 25,
  pixel_y_size = 25,
  asset_names = rsi::alos_palsar_band_mapping$planetary_computer_v1,
  stac_source = attr(asset_names, "stac_source"),
  collection = attr(asset_names, "collection_name"),
  query_function = attr(asset_names, "query_function"),
  download_function = attr(asset_names, "download_function"),
  sign_function = attr(asset_names, "sign_function"),
  rescale_bands = FALSE,
  item_filter_function = NULL,
  mask_band = attr(asset_names, "mask_band"),
  mask_function = attr(asset_names, "mask_function"),
  output_filename = paste0(proceduralnames::make_english_names(1), ".tif"),
  composite_function = "median",
  limit = 999,
  gdalwarp_options = rsi_gdalwarp_options(),
  gdal_config_options = rsi_gdal_config_options()
)

get_dem(
  aoi,
  ...,
  start_date = NULL,
  end_date = NULL,
  pixel_x_size = 30,
  pixel_y_size = 30,
  asset_names = rsi::dem_band_mapping$planetary_computer_v1$`cop-dem-glo-30`,
  stac_source = attr(asset_names, "stac_source"),
  collection = attr(asset_names, "collection_name"),
  query_function = attr(asset_names, "query_function"),
  download_function = attr(asset_names, "download_function"),
  sign_function = attr(asset_names, "sign_function"),
  rescale_bands = FALSE,
  item_filter_function = NULL,
  mask_band = NULL,
  mask_function = NULL,
  output_filename = paste0(proceduralnames::make_english_names(1), ".tif"),
  composite_function = "max",
  limit = 999,
  gdalwarp_options = rsi_gdalwarp_options(),
  gdal_config_options = rsi_gdal_config_options()
)
}
\arguments{
\item{aoi}{An sf(c) object outlining the area of interest to get imagery for.
Will be used to get the bounding box used for calculating metrics and the
output data's CRS.}

\item{start_date, end_date}{Character of length 1: The first and last date,
respectively, of imagery to include in metrics calculations. Should be in
YYYY-MM-DD format.}

\item{pixel_x_size, pixel_y_size}{Numeric of length 1: size of pixels in
x-direction (longitude / easting) and y-direction (latitude / northing).}

\item{asset_names}{The names of the assets to download. If this vector has
names, then the names of the vector are assumed to be the names of assets on
the STAC server, which will be renamed to the elements of the vector in the
final output.}

\item{stac_source}{Character of length 1: the STAC URL to download
imagery from.}

\item{collection}{Character of length 1: the STAC collection to download
images from.}

\item{...}{Passed to \code{item_filter_function}.}

\item{query_function}{A function that takes the output from
\code{\link[rstac:stac_search]{rstac::stac_search()}} and executes the request. See
\code{\link[=rsi_query_api]{rsi_query_api()}} and the \code{query_function} slots of
\link{sentinel1_band_mapping}, \link{sentinel2_band_mapping}, and
\link{landsat_band_mapping}.}

\item{download_function}{A function that takes the output from
\code{query_function} and downloads the assets attached to those items. See
\code{\link[=rsi_download_rasters]{rsi_download_rasters()}} for an example.}

\item{sign_function}{A function that takes the output from \code{query_function}
and signs the item URLs, if necessary.}

\item{rescale_bands}{Logical of length 1: If the STAC collection implements
the \code{raster} STAC extension, and that extension includes \code{scale} and \code{offset}
values, should this function attempt to automatically rescale the downloaded
data?}

\item{item_filter_function}{A function that takes the outputs of
\code{query_function} (usually a \code{STACItemCollection}) and \code{...} and returns a
filtered \code{STACItemCollection}. This is used, for example, to only download
images from specific Landsat platforms.}

\item{mask_band}{Character of length 1: The name of the asset in your
STAC source to use to mask the data. Set to \code{NULL} to not mask. See the
\code{mask_band} slots of \link{sentinel1_band_mapping}, \link{sentinel2_band_mapping}, and
\link{landsat_band_mapping}.}

\item{mask_function}{A function that takes a raster and returns a boolean
raster, where \code{TRUE} pixels will be preserved and \code{FALSE} or \code{NA} pixels will
be masked out. See \code{\link[=sentinel2_mask_function]{sentinel2_mask_function()}}.}

\item{output_filename}{The filename to write the output raster to. If
\code{composite_function} is \code{NULL}, item datetimes will be appended to this
in order to create unique filenames. If items do not have datetimes, a
sequential ID will be appended instead.}

\item{composite_function}{Character of length 1: The name of a
function used to combine downloaded images into a single composite
(i.e., to aggregate pixel values from multiple images into a single value).
Options include "merge", which 'stamps' images on top of one another such that
the "last" value downloaded for a pixel -- which isn't guaranteed to be the most
recent one -- will be the only value used, or any of "sum", "mean", "median",
"min", or "max", which consider all values available at each pixel.
Set to \code{NULL} to not composite
(i.e., to rescale and save each individual file independently).}

\item{limit}{an \code{integer} defining the maximum number of results
to return. If not informed, it defaults to the service implementation.}

\item{gdalwarp_options}{Options passed to \code{gdalwarp} through the \code{options}
argument of \code{\link[sf:gdal_utils]{sf::gdal_utils()}}. The same set of options are used for all
downloaded data and the final output images; this means that some common
options (for instance, \code{PREDICTOR=3}) may cause errors if bands are of
varying data types. The default values are provided by
\code{\link[=rsi_gdalwarp_options]{rsi_gdalwarp_options()}}.}

\item{gdal_config_options}{Options passed to \code{gdalwarp} through the
\code{config_options} argument of \code{\link[sf:gdal_utils]{sf::gdal_utils()}}. The default values are
provided by \code{\link[=rsi_gdal_config_options]{rsi_gdal_config_options()}}.}

\item{platforms}{The names of Landsat satellites to download imagery from.
These do not correspond to the \code{platforms} column in \code{\link[=spectral_indices]{spectral_indices()}};
the default argument of \code{c("landsat-9", "landsat-8")} corresponds to
the \code{Landsat-OLI} value in that column.}
}
\value{
\code{output_filename}, unchanged.
}
\description{
These functions retrieve raster data from STAC endpoints and optionally
create composite data sets from multiple files.
\code{get_stac_data()} is a generic function which should be able to download
raster data from a variety of data sources, while the other helper functions
have useful defaults for downloading common data sets from standard
STAC sources.
}
\section{Usage Tips}{

It's often useful to buffer your \code{aoi} object slightly, on the order of 1-2
cell widths, in order to ensure that data is downloaded for your entire AOI
even after accounting for any reprojection needed to compare your AOI to
the data on the STAC server.

These functions allow for parallelizing downloads via \code{\link[future:plan]{future::plan()}}, and
for user-controlled progress updates via \code{\link[progressr:handlers]{progressr::handlers()}}. If
there are fewer images to download than \code{asset_names}, then this function
uses \code{\link[=lapply]{lapply()}} to iterate through images and \code{\link[future.apply:future_mapply]{future.apply::future_mapply()}}
to iterate through downloading each asset. If there are more images than
assets, this function uses \code{\link[future.apply:future_lapply]{future.apply::future_lapply()}} to iterate through
images.
}

\section{Downloading from Planetary Computer}{


Certain data sets in Planetary Computer require
\href{https://planetarycomputer.microsoft.com/docs/concepts/sas/}{providing a subscription key}.
Even for non-protected data sets, providing a subscription key grants you
higher rate limits and faster downloads. As such, it's a good idea to
\href{https://planetarycomputer.microsoft.com/account/request}{request a Planetary Computer account},
then \href{https://planetarycomputer.developer.azure-api.net/}{generate a subscription key}.
If you set the \code{rsi_pc_key} environment variable to your key (either primary
or secondary; there is no difference), rsi will automatically use
this key to sign all requests against Planetary Computer.

There are currently some challenges with certain Landsat images in Planetary
Computer; please see
https://github.com/microsoft/PlanetaryComputer/discussions/101
for more information on these images and their current status. These files
may cause data downloads to fail.
}

\section{Compositing}{


This function can either download all data that intersects with your
spatiotemporal AOI as multiple files (if \code{composite_function = NULL}),
or can be used to rescale band values, apply a mask function, and create a
composite from the resulting files in a single function call. Each of these
steps can be skipped by passing \code{NULL} to the corresponding argument.

Masks are applied to each downloaded asset separately. Rescaling is applied
to the final composite after images are combined.

A number of the steps involved in creating composites -- rescaling band
values, running the mask function, masking images, and compositing images --
currently rely on the \code{terra} package for raster calculations. This means
creating larger composites, either in geographic or temporal dimension, may
cause errors. It can be a good idea to tile your \code{aoi} using
\code{sf::st_make_grid()} and iterate through the tiles to avoid these errors
(and to make it easier to interrupt and restart a download job).
}

\section{Rescaling}{

If \code{rescale_bands} is \code{TRUE}, then this function is able to use the \code{scale}
and \code{offset} values in the \code{bands} field of the \code{raster} STAC extension.
This was implemented originally to work with the Landsat collections in the
Planetary Computer STAC catalogue, but hopefully will work automatically with
other data sources as well. Note that Sentinel-2 data typically doesn't use
this STAC extension, and so the returned data is typically not re-scaled;
divide the downloaded band values by 10000 to get reflectance values in the
expected values.
}

\section{Sentinel-1 Data}{

The \code{get_sentinel1_imagery()} function is designed to download Sentinel-1 data
from the Microsoft Planetary Computer STAC API. Both the GRD and RTC
Sentinel-1 collections are supported. To download RTC data,
set \code{collection} to \code{sentinel-1-rtc}, and supply your subscription key
as an environment variable named \code{rsi_pc_key} (through, e.g., \code{Sys.setenv()}
or your \code{.Renviron} file).
}

\section{AlOS PALSAR Data}{

The \code{get_alos_palsar_imagery()} function is designed to download ALOS PALSAR
annual mosaic data from the Microsoft Planetary Computer STAC API. Data are
returned as a digital number (which is appropriate for some applications
and indices). To convert to backscatter (decibels) use the following formula:
\code{10 * log10(dn) - 83.0} where dn is the radar band in digital number.
}

\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
aoi <- sf::st_point(c(-74.912131, 44.080410))
aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)

get_stac_data(aoi,
  start_date = "2022-06-01",
  end_date = "2022-06-30",
  pixel_x_size = 30,
  pixel_y_size = 30,
  asset_names = c(
    "red", "blue", "green"
  ),
  stac_source = "https://planetarycomputer.microsoft.com/api/stac/v1/",
  collection = "landsat-c2-l2",
  query_function = rsi_query_api,
  sign_function = sign_planetary_computer,
  mask_band = "qa_pixel",
  mask_function = landsat_mask_function,
  item_filter_function = landsat_platform_filter,
  platforms = c("landsat-9", "landsat-8"),
  output_filename = tempfile(fileext = ".tif")
)

# or, mostly equivalently (will download more bands):
landsat_image <- get_landsat_imagery(
  aoi,
  start_date = "2022-06-01",
  end_date = "2022-08-30",
  output_filename = tempfile(fileext = ".tif")
)

landsat_image <- terra::rast(landsat_image)
terra::plotRGB(terra::stretch(landsat_image))

# The `get_*_imagery()` functions will download
# all available "data" assets by default
# (usually including measured values, and excluding derived bands)
sentinel1_data <- get_sentinel1_imagery(
  aoi,
  start_date = "2022-06-01",
  end_date = "2022-07-01",
  output_filename = tempfile(fileext = ".tif")
)
names(terra::rast(sentinel1_data))

# You can see what bands will be downloaded by a function
# by inspecting the corresponding `band_mapping` object:
sentinel2_band_mapping$planetary_computer_v1

# And you can add additional assets using `c()`:
c(
  sentinel2_band_mapping$planetary_computer_v1,
  "scl"
)

# Or subset the assets downloaded using `[` or `[[`:
sentinel2_imagery <- get_sentinel2_imagery(
  aoi,
  start_date = "2022-06-01",
  end_date = "2022-07-01",
  asset_names = sentinel2_band_mapping$planetary_computer_v1["B01"],
  output_filename = tempfile(fileext = ".tif")
)
names(terra::rast(sentinel2_imagery))

# If you're downloading data for a particularly large AOI,
# and can't composite the resulting images or want to make
# sure you can continue an interrupted download,
# consider tiling your AOI and requesting each tile separately:
aoi <- sf::st_make_grid(aoi, n = 2)
tiles <- lapply(
  seq_along(aoi),
  function(i) {
    get_landsat_imagery(
      aoi[i],
      start_date = "2022-06-01",
      end_date = "2022-08-30",
      output_filename = tempfile(fileext = ".tif")
    )
  }
)
# You'll get a list of tiles that you can then composite or
# work with however you wish:
unlist(tiles)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_indices.R
\name{calculate_indices}
\alias{calculate_indices}
\title{Calculate indices from the bands of a raster}
\usage{
calculate_indices(
  raster,
  indices,
  output_filename,
  ...,
  cores = 1L,
  wopt = list(),
  overwrite = FALSE,
  extra_objects = list(),
  names_suffix = NULL
)
}
\arguments{
\item{raster}{The raster (either as a SpatRaster or object readable by
\code{\link[terra:rast]{terra::rast()}}) to compute indices from.}

\item{indices}{A data frame of indices to compute. The intent is for this
function to work with subsets of \code{\link[=spectral_indices]{spectral_indices()}}, but you can provide
any data frame with \code{formula}, \code{bands}, and \code{short_name} columns; see
"Custom indices" below.}

\item{output_filename}{The filename to write the computed metrics to.}

\item{...}{These dots are for future extensions and must be empty.}

\item{cores}{positive integer. If \code{cores > 1}, a 'parallel' package cluster with that many cores is created and used}

\item{wopt}{list with named options for writing files as in \code{\link[terra]{writeRaster}}}

\item{overwrite}{logical. If \code{TRUE}, \code{filename} is overwritten}

\item{extra_objects}{A named list of additional objects to pass to the
minimal environment that formulas are executed in. For instance, if you
need to use the \code{pmax} function in order to calculate an index, you can
make it available in the environment by setting
\code{extra_objects = list("pmax" = pmax)}. Providing extra functionality is
inherently less safe than the default minimal environment, and as such always
emits a warning, which you can suppress with \code{suppressWarnings()}.}

\item{names_suffix}{If not \code{NULL}, will be used (with \code{\link[=paste]{paste()}}) to add a
suffix to each of the band names returned.}
}
\value{
\code{output_filename}, unchanged.
}
\description{
This function computes any number of indices from an input raster via
\code{\link[terra:predict]{terra::predict()}}. By default, this function is designed to work with
subsets of \code{\link[=spectral_indices]{spectral_indices()}}, but it will work with any data frame with a
\code{formula}, \code{bands}, and \code{short_name} column.
}
\section{Security}{

Note that this function is running code from the \code{formula} column of the
spectral indices data frame, which is derived from a JSON file downloaded off
the internet. It's not impossible that an attacker could take advantage of
this to run arbitrary code on your computer. To mitigate this, indices are
calculated in a minimal environment that contains very few functions or
symbols (preventing an attacker from accessing, for example, \code{system()}).

Still, it's good practice to inspect your \code{formula} column to make sure
there's nothing nasty hiding in any of the formulas you're going to run.
Additionally, consider using pre-saved indices tables or
\code{spectral_indices(download_indices = FALSE)} if using this in an unsupervised
workload.
}

\section{Custom indices}{

While this function is designed to work with subsets of \code{\link[=spectral_indices]{spectral_indices()}},
you can also provide your own custom indices by providing a data frame with
at least three columns:
\itemize{
\item \code{formula} should contain a string representation of the equation used to
calculate the index. This string will be interpreted as R code, and should use
normal R arithmetic operators. Any additional functions used in calculating the
index will need to be passed through \code{extra_objects}.
\item \code{bands} should be a list column containing character vectors identifying the
bands required to calculate the index. These vectors will be checked against the
band names of \code{raster} to make sure that all necessary bands are available
before indices are calculated.
\item \code{short_name} should contain a string which will be used as the band name in
the final output file.
}

Sometimes you might want to tweak the formula used for an existing index, for
instance to set non-data parameters to constant values. In that case, make sure
you also update \code{bands} to reflect just the bands required by your updated formula,
so that the function won't error unnecessarily.
}

\examples{
our_raster <- system.file("rasters/example_sentinel1.tif", package = "rsi")
calculate_indices(
  our_raster,
  filter_bands(bands = names(terra::rast(our_raster))),
  tempfile(fileext = ".tif"),
  names_suffix = "sentinel1"
)

# Formulas aren't able to access most R functions or operators,
# in order to try and keep formulas from doing something bad:
example_indices <- filter_platforms(platforms = "Sentinel-1 (Dual Polarisation VV-VH)")[1, ]
example_indices$formula <- 'system("echo something bad")'
# So this will error:
try(
  calculate_indices(
    system.file("rasters/example_sentinel1.tif", package = "rsi"),
    example_indices,
    tempfile(fileext = ".tif")
  )
)

# Because of this, formulas which try to use most R functions
# will wind up erroring as well:
example_indices$formula <- "pmax(VH, VV)"
try(
  calculate_indices(
    system.file("rasters/example_sentinel1.tif", package = "rsi"),
    example_indices,
    tempfile(fileext = ".tif")
  )
)

# To fix this, pass the objects you want to use to `extra_objects`
calculate_indices(
  system.file("rasters/example_sentinel1.tif", package = "rsi"),
  example_indices,
  tempfile(fileext = ".tif"),
  extra_objects = list(pmax = pmax)
) |>
  suppressWarnings(classes = "rsi_extra_objects")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{filter_platforms}
\alias{filter_platforms}
\alias{filter_bands}
\title{Filter indices based on (relatively) complicated fields}
\usage{
filter_platforms(
  indices = spectral_indices(),
  platforms = unique(unlist(spectral_indices(download_indices = FALSE, update_cache =
    FALSE)$platforms)),
  operand = c("all", "any")
)

filter_bands(
  indices = spectral_indices(),
  bands = unique(unlist(spectral_indices(download_indices = FALSE, update_cache =
    FALSE)$bands)),
  operand = c("all", "any"),
  type = c("filter", "search")
)
}
\arguments{
\item{indices}{The data frame to filter. Must contain the relevant column.}

\item{platforms, bands}{Names of the instruments (for \code{platforms}) or spectra
(for \code{bands}) indices must contain.}

\item{operand}{A function defining how to apply this filter.
For instance, \code{operand = all} means that the index must contain all the
\code{platforms} or \code{bands} provided, while \code{operand = any} means that the index
must contain at least one of the \code{platforms} or \code{bands} provided.}

\item{type}{What type of query is this? If \code{filter}, then indices are
returned if all/any the bands they use (depending on \code{operand}) are in
\code{bands}. If \code{search}, then indices are returned if all/any of \code{bands} are in
the bands they use.}
}
\value{
A filtered version of \code{indices}.
}
\description{
Filter indices based on (relatively) complicated fields
}
\examples{
filter_platforms(platforms = "Sentinel-2")
filter_platforms(platforms = c("Landsat-OLI", "Sentinel-2"))
filter_bands(bands = c("R", "N"), operand = any)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spectral_indices.R
\name{spectral_indices_url}
\alias{spectral_indices_url}
\title{Get the URL to download spectral indices from}
\usage{
spectral_indices_url()
}
\value{
A URL to download indices from.
}
\description{
Get the URL to download spectral indices from
}
\examples{
spectral_indices_url()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download.R
\name{rsi_download_rasters}
\alias{rsi_download_rasters}
\title{Download specific assets from a set of STAC items}
\usage{
rsi_download_rasters(
  items,
  aoi,
  asset_names,
  sign_function = NULL,
  merge = FALSE,
  gdalwarp_options = c("-r", "bilinear", "-multi", "-overwrite", "-co",
    "COMPRESS=DEFLATE", "-co", "PREDICTOR=2", "-co", "NUM_THREADS=ALL_CPUS"),
  gdal_config_options = c(VSI_CACHE = "TRUE", GDAL_CACHEMAX = "30\%", VSI_CACHE_SIZE =
    "10000000", GDAL_HTTP_MULTIPLEX = "YES", GDAL_INGESTED_BYTES_AT_OPEN = "32000",
    GDAL_DISABLE_READDIR_ON_OPEN = "EMPTY_DIR", GDAL_HTTP_VERSION = "2",
    GDAL_HTTP_MERGE_CONSECUTIVE_RANGES = "YES", GDAL_NUM_THREADS = "ALL_CPUS",
    GDAL_HTTP_USERAGENT = "rsi (https://permian-global-research.github.io/rsi/)"),
  ...
)
}
\arguments{
\item{items}{A \code{StacItemCollection} object, as returned by \code{\link[=rsi_query_api]{rsi_query_api()}}.}

\item{aoi}{Either an sf(c) object outlining the area of interest to get
imagery for, or a \code{bbox} image containing the bounding box of your AOI.}

\item{asset_names}{The names of the assets to download. If this vector has
names, then the names of the vector are assumed to be the names of assets on
the STAC server, which will be renamed to the elements of the vector in the
final output.}

\item{sign_function}{A function that takes the output from \code{query_function}
and signs the item URLs, if necessary.}

\item{merge}{Logical: for each asset, should data from multiple items be
merged into a single downloaded file? If \code{TRUE}, this returns a single file
for each asset, which has been merged via gdalwarp. No resampling or
compositing is performed, but rather each pixel uses the last data
downloaded. This is fast, but precludes per-item masking and compositing.
If \code{FALSE}, each asset from each item is saved as a separate file.}

\item{gdalwarp_options}{Options passed to \code{gdalwarp} through the \code{options}
argument of \code{\link[sf:gdal_utils]{sf::gdal_utils()}}. The same set of options are used for all
downloaded data and the final output images; this means that some common
options (for instance, \code{PREDICTOR=3}) may cause errors if bands are of
varying data types. The default values are provided by
\code{\link[=rsi_gdalwarp_options]{rsi_gdalwarp_options()}}.}

\item{gdal_config_options}{Options passed to \code{gdalwarp} through the
\code{config_options} argument of \code{\link[sf:gdal_utils]{sf::gdal_utils()}}. The default values are
provided by \code{\link[=rsi_gdal_config_options]{rsi_gdal_config_options()}}.}

\item{...}{Passed to \code{item_filter_function}.}
}
\value{
A data frame  where columns correspond to distinct assets, rows
correspond to distinct items, and cells contain file paths to the downloaded
data.
}
\description{
Download specific assets from a set of STAC items
}
\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
aoi <- sf::st_point(c(-74.912131, 44.080410))
aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)

landsat_image <- get_landsat_imagery(
  aoi,
  start_date = "2022-06-01",
  end_date = "2022-08-30",
  download_function = rsi_download_rasters
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{landsat_band_mapping}
\alias{landsat_band_mapping}
\title{Landsat band mapping}
\format{
An object of class \code{list} of length 1.
}
\usage{
landsat_band_mapping
}
\description{
This object is a named list of character vectors, with names corresponding to
Landsat band names and values corresponding to band names in
\code{spectral_indices}.
}
\details{
Band mapping objects:

These objects are semi-standardized sets of metadata which provide all the
necessary information for downloading data from a given STAC server.
The object itself is list of character vectors, whose names represent asset
names on a given STAC server and whose values represent the corresponding
standardized band name from the Awesome Spectral Indices project. In addition
to this data, these vectors usually have some of (but not necessarily all of)
the following attributes:
\itemize{
\item \code{stac_source}: The URL for the STAC server this metadata corresponds to.
\item \code{collection_name}: The default STAC collection for this data source.
\item \code{download_function}: The function to be used to download assets from the
STAC server.
\item \code{mask_band}: The name of the asset on this server to be used for masking
images.
\item \code{mask_function}: The function to be used to mask images downloaded from
this server.
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rsi-package.R
\docType{package}
\name{rsi-package}
\alias{rsi}
\alias{rsi-package}
\title{rsi: Efficiently Retrieve and Process Satellite Imagery}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Downloads spatial data from spatiotemporal asset catalogs ('STAC'), computes standard spectral indices from the Awesome Spectral Indices project (Montero et al. (2023) \doi{10.1038/s41597-023-02096-0}) against raster data, and glues the outputs together into predictor bricks. Methods focus on interoperability with the broader spatial ecosystem; function arguments and outputs use classes from 'sf' and 'terra', and data downloading functions support complex 'CQL2' queries using 'rstac'.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/Permian-Global-Research/rsi}
  \item \url{https://permian-global-research.github.io/rsi/}
  \item Report bugs at \url{https://github.com/Permian-Global-Research/rsi/issues}
}

}
\author{
\strong{Maintainer}: Michael Mahoney \email{mike.mahoney.218@gmail.com} (\href{https://orcid.org/0000-0003-2402-304X}{ORCID})

Other contributors:
\itemize{
  \item Felipe Carvalho (Felipe reviewed the package (v. 0.3.0) for rOpenSci, see <https://github.com/ropensci/software-review/issues/636>) [reviewer]
  \item Michael Sumner (Michael reviewed the package (v. 0.3.0) for rOpenSci, see <https://github.com/ropensci/software-review/issues/636>) [reviewer]
  \item Permian Global [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mask_functions.R
\name{landsat_mask_function}
\alias{landsat_mask_function}
\title{Create a Landsat mask raster from the QA band}
\usage{
landsat_mask_function(
  raster,
  include = c("land", "water", "both"),
  ...,
  masked_bits
)
}
\arguments{
\item{raster}{The QA band of a Landsat image}

\item{include}{Include pixels that represent land, water, or both? Passing
\code{c("land", "water")} is identical to passing \code{"both"}.}

\item{...}{These dots are for future extensions and must be empty.}

\item{masked_bits}{Optionally, a list of integer vectors representing the
individual bits to mask out. Each vector is converted to an integer
representation, and then pixels with matching \code{qa_pixel} values
are preserved by the mask. Refer to the Landsat science product guide for
further information on what bit values represent for your platform of
interest.}
}
\value{
A boolean raster to be used to mask a Landsat image
}
\description{
Create a Landsat mask raster from the QA band
}
\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
aoi <- sf::st_point(c(-74.912131, 44.080410))
aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)

landsat_image <- get_landsat_imagery(
  aoi,
  start_date = "2022-06-01",
  end_date = "2022-08-30",
  mask_function = landsat_mask_function,
  output_file = tempfile(fileext = ".tif")
)

# Or, optionally pass the qa_pixel bits to mask out directly
landsat_image <- get_landsat_imagery(
  aoi,
  start_date = "2022-06-01",
  end_date = "2022-08-30",
  mask_function = \(x) landsat_mask_function(
    x,
    masked_bits = list(c(0:5, 7, 9, 11, 13, 15))
  ),
  output_file = tempfile(fileext = ".tif")
)

# You can use this to specify multiple acceptable values
# from the qa_pixel bitmask; names are optional
landsat_image <- get_landsat_imagery(
  aoi,
  start_date = "2022-06-01",
  end_date = "2022-08-30",
  mask_function = \(x) landsat_mask_function(
    x,
    masked_bits = list(
      clear_land = c(0:5, 7, 9, 11, 13, 15),
      clear_water = c(0:5, 9, 11, 13, 15)
    )
  ),
  output_file = tempfile(fileext = ".tif")
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stack_rasters.R
\name{stack_rasters}
\alias{stack_rasters}
\title{Create and save a multi-band output raster by combining input rasters}
\usage{
stack_rasters(
  rasters,
  output_filename,
  ...,
  resolution,
  extent,
  reference_raster = 1,
  resampling_method = "bilinear",
  band_names,
  check_crs = TRUE,
  gdalwarp_options = c("-multi", "-overwrite", "-co", "COMPRESS=DEFLATE", "-co",
    "PREDICTOR=2", "-co", "NUM_THREADS=ALL_CPUS"),
  gdal_config_options = c(VSI_CACHE = "TRUE", GDAL_CACHEMAX = "30\%", VSI_CACHE_SIZE =
    "10000000", GDAL_HTTP_MULTIPLEX = "YES", GDAL_INGESTED_BYTES_AT_OPEN = "32000",
    GDAL_DISABLE_READDIR_ON_OPEN = "EMPTY_DIR", GDAL_HTTP_VERSION = "2",
    GDAL_HTTP_MERGE_CONSECUTIVE_RANGES = "YES", GDAL_NUM_THREADS = "ALL_CPUS")
)
}
\arguments{
\item{rasters}{A list of rasters to combine into a single multi-band raster,
as character file paths to files that can be read by \code{\link[terra:rast]{terra::rast()}}.
Rasters will be "stacked" upon one another, preserving values. They must
share CRS.}

\item{output_filename}{The location to save the final "stacked" raster. If
this filename has a "vrt" extension as determined by \code{tools::file_ext()},
then this function exits after creating a VRT; otherwise, this function will
create a VRT and then use \code{sf::gdal_utils("warp")} to convert the VRT into
another format.}

\item{...}{These dots are for future extensions and must be empty.}

\item{resolution}{Numeric of length 2, representing the target X and Y
resolution of the output raster. If only a single value is provided, it will
be used for both X and Y resolution; if more than 2 values are provided, an
error is thrown.}

\item{extent}{Numeric of length 4, representing the target xmin, ymin, xmax,
and ymax values of the output raster (its bounding box), in that order.}

\item{reference_raster}{The position (index) of the raster in \code{rasters} to
take extent, resolution, and CRS information from. No reprojection is done.
If \code{resolution} or \code{extent} are provided, they override the values from the
reference raster.}

\item{resampling_method}{The method to use when resampling to different
resolutions in the VRT.}

\item{band_names}{Either a character vector of band names, or a function that
when given a character vector of band names, returns a character vector of
the same length containing new band names.}

\item{check_crs}{Logical: Should this function check that all \code{rasters} share
the same CRS? Set to \code{FALSE} only if you are entirely confident that rasters
have equivalent CRS definitions, but not identical WKT strings.}

\item{gdalwarp_options}{Options passed to \code{gdalwarp} through the \code{options}
argument of \code{\link[sf:gdal_utils]{sf::gdal_utils()}}. This argument is ignored (with a warning)
if \code{output_filename} is a VRT.}

\item{gdal_config_options}{Options passed to \code{gdalwarp} through the
\code{config_options} argument of \code{\link[sf:gdal_utils]{sf::gdal_utils()}}.  This argument is ignored
(with a warning) if \code{output_filename} is a VRT.}
}
\value{
\code{output_filename}, unchanged.
}
\description{
This function creates an output raster that "stacks" all the bands of its
input rasters, as though they were loaded one after another into a GIS. It
does this by first constructing a GDAL virtual raster, or "VRT", and then
optionally uses GDAL's warper to convert this VRT into a standalone file.
The VRT is fast to create and does not require much space, but does require
the input rasters not be moved or altered. Creating a standalone raster from
this file may take a long time and a large amount of disk space.
}
\examples{
stack_rasters(
  list(
    system.file("rasters/dpdd.tif", package = "rsi"),
    system.file("rasters/example_sentinel1.tif", package = "rsi")
  ),
  tempfile(fileext = ".vrt")
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query_and_sign.R
\name{sign_planetary_computer}
\alias{sign_planetary_computer}
\title{Sign STAC items retrieved from the Planetary Computer}
\usage{
sign_planetary_computer(items, subscription_key = Sys.getenv("rsi_pc_key"))
}
\arguments{
\item{items}{A STACItemCollection, as returned by \code{rsi_query_api}.}

\item{subscription_key}{Optionally, a subscription key associated with your
Planetary Computer account. At the time of writing, this is required for
downloading Sentinel 1 RTC products, as well as NAIP imagery. This key will
be automatically used if the environment variable \code{rsi_pc_key} is set.}
}
\value{
A STACItemCollection object with signed assets url.
}
\description{
Sign STAC items retrieved from the Planetary Computer
}
\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
aoi <- sf::st_point(c(-74.912131, 44.080410))
aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)

landsat_image <- get_landsat_imagery(
  aoi,
  start_date = "2022-06-01",
  end_date = "2022-08-30",
  sign_function = sign_planetary_computer
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query_and_sign.R
\name{rsi_query_api}
\alias{rsi_query_api}
\title{Query a STAC API using a specific spatiotemporal area of interest}
\usage{
rsi_query_api(bbox, stac_source, collection, start_date, end_date, limit, ...)
}
\arguments{
\item{bbox}{A \code{bbox} or \code{sfc} object, from the sf package, representing the
spatial bounding box of your area of interest. This must be in EPSG:4326
coordinates (and, if this function is called from within \code{get_stac_data()},
it will be) or else it will be automatically reprojected.}

\item{stac_source}{Character of length 1: the STAC URL to download
imagery from.}

\item{collection}{Character of length 1: the STAC collection to download
images from.}

\item{start_date, end_date}{Character strings of length 1 representing the
boundaries of your temporal range of interest, in RFC-3339 format. Set either
argument to \code{..} to use an open interval; set \code{start_date} to \code{NULL} to not
pass a temporal range of interest (which may cause errors with some APIs). If
this function is called from within \code{get_stac_data()}, the inputs to
\code{start_date} and \code{end_date} will have already been processed to try and force
RFC-3339 compliance.}

\item{limit}{an \code{integer} defining the maximum number of results
to return. If not informed, it defaults to the service implementation.}

\item{...}{Ignored by this function. Arguments passed to \code{get_stac_data()}
via \code{...} will be available (unchanged) in this function}
}
\value{
A StacItemCollection object.
}
\description{
This function is the default method used to retrieve lists of items to
download for all the collections and endpoints supported by rsi. It will
likely work for any other STAC APIs of interest.
}
\details{
You can pass your own query functions to \code{get_stac_data()} and its variants.
This is the best way to perform more complex queries, for instance if you
need to provide authentication to get the list of items (not just the assets)
available for your AOI, or to perform cloud filtering prior to downloading
assets.
}
\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
aoi <- sf::st_point(c(-74.912131, 44.080410))
aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)

landsat_image <- get_landsat_imagery(
  aoi,
  start_date = "2022-06-01",
  end_date = "2022-08-30",
  query_function = rsi_query_api
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mask_functions.R
\name{alos_palsar_mask_function}
\alias{alos_palsar_mask_function}
\title{Create an ALOS PALSAR mask raster from the mask band}
\usage{
alos_palsar_mask_function(raster, include = c("land", "water", "both"))
}
\arguments{
\item{raster}{The mask band of an ALOS PALSAR image}

\item{include}{Include pixels that represent land, water, or both? Passing
\code{c("land", "water")} is identical to passing \code{"both"}.}
}
\value{
A boolean raster to be used to mask an ALOS PALSAR image
}
\description{
Create an ALOS PALSAR mask raster from the mask band
}
\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
aoi <- sf::st_point(c(-74.912131, 44.080410))
aoi <- sf::st_set_crs(sf::st_sfc(aoi), 4326)
aoi <- sf::st_buffer(sf::st_transform(aoi, 5070), 100)

palsar_image <- get_alos_palsar_imagery(
  aoi,
  start_date = "2021-01-01",
  end_date = "2021-12-31",
  mask_function = alos_palsar_mask_function,
  output_file = tempfile(fileext = ".tif"),
  gdalwarp_options = c(
    rsi::rsi_gdalwarp_options(),
    "-srcnodata", "nan"
  )
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{deprecated}
\alias{deprecated}
\alias{default_query_function}
\title{Deprecated functions}
\usage{
default_query_function(
  bbox,
  stac_source,
  collection,
  start_date,
  end_date,
  limit,
  ...
)
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

These functions have been deprecated in favor of better approaches.
\itemize{
\item \code{default_query_function()} was renamed to \code{rsi_query_api()}. These
functions are identical, and the older name will be removed in a future
release.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spectral_indices.R
\name{spectral_indices}
\alias{spectral_indices}
\title{Get a data frame of spectral indices}
\source{
\url{https://github.com/awesome-spectral-indices/awesome-spectral-indices}
}
\usage{
spectral_indices(
  ...,
  url = spectral_indices_url(),
  download_indices = NULL,
  update_cache = NULL
)
}
\arguments{
\item{...}{These dots are for future extensions and must be empty.}

\item{url}{The URL to download spectral indices from. If the option \code{rsi_url}
is set, that value will be used; otherwise, if the environment variable
\code{rsi_url} is set, that value will be used; otherwise, the list at
https://github.com/awesome-spectral-indices/awesome-spectral-indices will
be used.}

\item{download_indices}{Logical: should this function download indices? If
\code{NULL}, this function will only download indices if the cache will be
updated. If \code{TRUE}, this function will attempt to download indices no matter
what. If \code{FALSE}, either cached or package indices will be used.}

\item{update_cache}{Logical: should cached indices be updated? If \code{NULL},
cached values will be updated if the cache is older than a day. If \code{TRUE},
the cache will be updated, if \code{FALSE} it will not.}
}
\value{
A \link[tibble:tibble]{tibble::tibble} with nine columns, containing information about spectral indices.
}
\description{
This function returns a data frame of spectral indices, from the
\code{awesome-spectral-indices} repository.
}
\examples{
spectral_indices()

}
